#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/3/8 12:23
@File    : orm_stg_run_status_detail_test.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_utils.mess import str_2_datetime
from ibats_common.backend.orm import *
from datetime import date, datetime, time
import unittest
from ibats_common.test.orm_init_test import InitTest
from ibats_common.common import ExchangeName
from ibats_common.test.orm_pos_status_detail_test import PosStatusDetailTest


class TradeAgentStatusDetailTest(unittest.TestCase):  # 继承unittest.TestCase
    def tearDown(self):
        pass

    def setUp(self):
        # 每个测试用例执行之前做操作
        pass

    @classmethod
    def tearDownClass(cls):
        # 必须使用 @ classmethod装饰器, 所有test运行完后运行一次
        pass

    @classmethod
    def setUpClass(cls):
        # from ibats_common.config import update_db_config, ConfigBase
        # update_db_config({
        #     ConfigBase.DB_SCHEMA_IBATS: 'mysql://mg:Dcba1234@localhost/' + ConfigBase.DB_SCHEMA_IBATS,
        # })
        # 必须使用@classmethod 装饰器,所有test运行前运行一次
        init()
        global engine_ibats
        engine_ibats = engines.engine_ibats
        config.ORM_UPDATE_OR_INSERT_PER_ACTION = True

    def test_create(self):
        _, status = TradeAgentStatusDetailTest.add_trade_agent_status_detail()
        with with_db_session(engine_ibats) as session:
            status2 = session.query(TradeAgentStatusDetail).filter(
                TradeAgentStatusDetail.trade_agent_status_detail_idx == status.trade_agent_status_detail_idx,
                TradeAgentStatusDetail.stg_run_id == status.stg_run_id
            ).first()
        self.assertIsInstance(status2, TradeAgentStatusDetail)
        self.assertGreater(status2.trade_agent_status_detail_idx, -1)
        self.assertGreater(status2.cash_init, 0)
        self.assertEqual(status2.cash_init, status2.cash_available)
        self.assertEqual(status2.cash_init, status2.cash_and_margin)
        self.assertEqual(status2.commission_tot, 0)
        self.assertEqual(status2.floating_pl_cum, 0)
        self.assertEqual(status2.position_profit, 0)
        self.assertEqual(status2.close_profit, 0)
        self.assertEqual(status2.curr_margin, 0)
        self.assertIsInstance(status2.trade_dt, datetime)
        self.assertIsInstance(status2.trade_time, time)
        self.assertIsInstance(status2.trade_millisec, int)

    def test_update_by_pos_status_detail(self):
        info, status = TradeAgentStatusDetailTest.add_trade_agent_status_detail()
        pos_status = TradeAgentStatusDetailTest.add_pos_status_detail(info)
        timestamp_curr = pd.Timestamp(str_2_datetime('2018-12-15 13:24:35'))
        status2 = status.update_by_pos_status_detail({pos_status.symbol: pos_status}, timestamp_curr=timestamp_curr)
        self.assertIsInstance(status2, TradeAgentStatusDetail)
        self.assertEqual(status2.stg_run_id, info.stg_run_id)
        self.assertGreater(status2.trade_agent_status_detail_idx, status.trade_agent_status_detail_idx)
        self.assertEqual(status2.cash_available, status2.cash_init - pos_status.margin - pos_status.commission)
        self.assertEqual(status2.position_value, pos_status.position_value)
        self.assertEqual(status2.curr_margin, pos_status.margin)
        self.assertEqual(status2.close_profit, 0)
        self.assertEqual(status2.position_profit, pos_status.floating_pl)
        self.assertEqual(status2.floating_pl_cum, pos_status.floating_pl_cum)
        self.assertEqual(status2.commission_tot, pos_status.commission)
        self.assertEqual(status2.cash_init, status.cash_init)
        self.assertEqual(status2.cash_and_margin, status2.cash_available + status2.curr_margin)
        self.assertEqual(status2.cashflow_daily, - pos_status.margin - pos_status.commission)
        self.assertEqual(status2.cashflow_cum, status.cashflow_cum + status2.cashflow_daily)

    @staticmethod
    def add_trade_agent_status_detail():
        info = InitTest.add_stg_run_info()
        init_cash = 1000000
        timestamp_curr = pd.Timestamp(str_2_datetime('2018-12-14 13:24:35'))
        status = TradeAgentStatusDetail.create_t_1(info.stg_run_id, ExchangeName.DataIntegration, init_cash, timestamp_curr=timestamp_curr)
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(status)
            session.commit()
        return info, status

    @staticmethod
    def add_pos_status_detail(info):
        trade = PosStatusDetailTest.add_trade(info.stg_run_id)
        pos_status = PosStatusDetail.create_by_trade_detail(trade)
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(pos_status)
            session.commit()
        return pos_status


if __name__ == '__main__':
    unittest.main()  # 运行所有的测试用例



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/11/7 10:42
@File    : ma_cross_stg.py
@contact : mmmaaaggg@163.com
@desc    : 简单的 MA5、MA10金叉、死叉多空策略，仅供测试及演示使用
"""
from ibats_common.common import BacktestTradeMode, ContextKey, Direction, CalcMode
from ibats_common.strategy import StgBase
from ibats_common.strategy_handler import strategy_handler_factory
from ibats_local_trader.agent.md_agent import *
from ibats_local_trader.agent.td_agent import *

logger = logging.getLogger(__name__)


class MACrossStg(StgBase):

    def __init__(self, unit=1):
        super().__init__()
        self.unit = unit
        self.ma5 = []
        self.ma10 = []

    def on_prepare_min1(self, md_df, context):
        if md_df is not None:
            self.ma5 = list(md_df['close'].rolling(5, 5).mean())[10:]
            self.ma10 = list(md_df['close'].rolling(10, 10).mean())[10:]

    def on_min1(self, md_df, context):
        close = md_df['close'].iloc[-1]
        self.ma5.append(md_df['close'].iloc[-5:].mean())
        self.ma10.append(md_df['close'].iloc[-10:].mean())
        instrument_id = context[ContextKey.instrument_id_list][0]
        if self.ma5[-2] < self.ma10[-2] and self.ma5[-1] > self.ma10[-1]:
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Short:
                        self.close_short(instrument_id, close, pos_info.position)
                    elif direction == Direction.Long:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_long(instrument_id, close, self.unit)
        elif self.ma5[-2] > self.ma10[-2] and self.ma5[-1] < self.ma10[-1]:
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Long:
                        self.close_long(instrument_id, close, pos_info.position)
                    elif direction == Direction.Short:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_short(instrument_id, close, self.unit)


def _test_use(is_plot):
    from ibats_common import module_root_path
    import os
    # 参数设置
    run_mode = RunMode.Backtest
    trade_mode = BacktestTradeMode.Order_2_Deal
    calc_mode = CalcMode.Normal
    md_agent_params_list = [{
        'md_period': PeriodType.Min1,
        'instrument_id_list': ['RB'],
        'datetime_key': 'trade_date',
        'init_md_date_from': '1995-1-1',  # 行情初始化加载历史数据，供策略分析预加载使用
        'init_md_date_to': '2010-1-1',
        # 'C:\GitHub\IBATS_Common\ibats_common\example\ru_price2.csv'
        'file_path': os.path.abspath(os.path.join(module_root_path, 'example', 'data', 'RB.csv')),
        'symbol_key': 'instrument_type',
    }]
    if run_mode == RunMode.Realtime:
        strategy_params = {'unit': 100}
        trade_agent_params = {
        }
        strategy_handler_param = {
        }
    elif run_mode == RunMode.Backtest:
        strategy_params = {'unit': 100}
        trade_agent_params = {
            'trade_mode': trade_mode,
            'init_cash': 1000000,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2010-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    else:
        # RunMode.Backtest_FixPercent
        strategy_params = {'unit': 1}
        trade_agent_params = {
            'trade_mode': trade_mode,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2010-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }

    # 初始化策略处理器
    stghandler = strategy_handler_factory(
        stg_class=MACrossStg,
        strategy_params=strategy_params,
        md_agent_params_list=md_agent_params_list,
        exchange_name=ExchangeName.LocalFile,
        run_mode=run_mode,
        trade_agent_params=trade_agent_params,
        strategy_handler_param=strategy_handler_param,
    )
    stghandler.start()
    time.sleep(10)
    stghandler.keep_running = False
    stghandler.join()
    stg_run_id = stghandler.stg_run_id
    logging.info("执行结束 stg_run_id = %d", stg_run_id)

    if is_plot:
        # from ibats_common.analysis.plot_db import show_order, show_cash_and_margin, show_rr_with_md
        # from ibats_common.analysis.summary import summary_rr
        # show_order(stg_run_id)
        # df = show_cash_and_margin(stg_run_id)
        # sum_df, symbol_rr_dic, save_file_path_dic = show_rr_with_md(stg_run_id)
        # for symbol, rr_df in symbol_rr_dic.items():
        #     col_transfer_dic = {'return': rr_df.columns}
        #     summary_rr(rr_df, figure_4_each_col=True, col_transfer_dic=col_transfer_dic)
        from ibats_common.analysis.summary import summary_stg_2_docx
        from ibats_utils.mess import open_file_with_system_app
        file_path = summary_stg_2_docx(stg_run_id)
        open_file_with_system_app(file_path)

    return stg_run_id


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG, format=config.LOG_FORMAT)
    is_plot = True
    _test_use(is_plot)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/4/7 16:31
@File    : plot.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import itertools
import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from ibats_utils.mess import date_2_str
from scipy import stats

from ibats_common.analysis import get_cache_folder_path
from ibats_common.backend.label import calc_label2, calc_label3

logger = logging.getLogger(__name__)


def get_file_name(header, name=None):
    file_name = f'{header} {np.random.randint(10000)}.png' if name is None else f'{header} {name}.png'
    return file_name


def clean_cache():
    """
    清空cache目录
    :return:
    """
    folder_path = get_cache_folder_path()
    for root, dirs, files in os.walk(folder_path, topdown=False):
        # print(root, dirs, files)
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))


def drawdown_plot(df: pd.DataFrame, perf_stats=None, col_name_list=None,
                  enable_show_plot=True, enable_save_plot=False, name=None):
    """
    回撤曲线
    :param df:
    :param perf_stats:
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if col_name_list is None:
        data_df = df.to_drawdown_series()
        col_name_list = list(df.columns)
    else:
        data_df = df[col_name_list].to_drawdown_series()

    if perf_stats is None:
        perf_stats = df.calc_stats()

    col_mdd_len_dic = {col_name: s.drawdown_details.Length.max()
                       for col_name, s in perf_stats.items() if s.drawdown_details is not None}
    col_mdd_dic = {col_name: mdd for col_name, mdd in data_df.min().items()}
    data_df.rename(
        columns={col_name: f"{col_name}[{mdd * 100:.2f}% {col_mdd_len_dic.setdefault(col_name, '')}]"
                 for col_name, mdd in col_mdd_dic.items()},
        inplace=True)

    ax = data_df.plot()
    ax.set_title(f"Drawdown {['{:.2f}%'.format(col_mdd_dic[_] * 100) for _ in col_name_list]}")

    if enable_save_plot:
        file_name = get_file_name(f'drawdown', name=name)
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    if enable_show_plot:
        plt.show()

    plt.cla()
    plt.clf()

    return data_df, file_path


def plot_corr(df: pd.DataFrame, perf_stats=None,
              enable_show_plot=True, enable_save_plot=False, name=None):
    """
    相关性矩阵图
    :param df:
    :param perf_stats:
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if perf_stats is None:
        perf_stats = df.calc_stats()

    ax = perf_stats.plot_correlation()
    plt.suptitle("Correlation")

    if enable_save_plot:
        file_name = get_file_name(f'correlation', name=name)
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    if enable_show_plot:
        plt.show()

    plt.cla()
    plt.clf()

    return file_path


def plot_rr_df(df: pd.DataFrame, col_name_list=None, enable_show_plot=True, enable_save_plot=False, name=None):
    """
    Return Rate
    :param df:
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if col_name_list is not None:
        data_df = df[col_name_list]
    else:
        data_df = df

    # """绘图函数"""
    ax = data_df.plot()
    ax.set_title(
        f"Return Rate " if name is None else f"Return Rate [{name}] "
        f"{date_2_str(min(data_df.index))} - {date_2_str(max(data_df.index))} ({data_df.shape[0]} days)")

    if enable_save_plot:
        file_name = get_file_name(f'rr', name=name)
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    if enable_show_plot:
        plt.show()

    plt.cla()
    plt.clf()

    return file_path


def plot_scatter_matrix(df: pd.DataFrame, diagonal='hist', col_name_list=None, enable_show_plot=True,
                        enable_save_plot=False, name=None):
    """
    plot scatter_matrix
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.plotting.scatter_matrix.html
    :param df:
    :param diagonal: diagonal，必须且只能在{'hist', 'kde'}中选择1个，
    'hist'表示直方图(Histogram plot),'kde'表示核密度估计(Kernel Density Estimation)
    该参数是scatter_matrix函数的关键参数
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if col_name_list is not None:
        data_df = df[col_name_list]
    else:
        data_df = df

    # """绘图函数"""
    pd.plotting.scatter_matrix(data_df)
    plt.suptitle(
        f"Scatter Matrix " if name is None else f"Scatter Matrix [{name}] "
        f"{date_2_str(min(data_df.index))} - {date_2_str(max(data_df.index))} ({data_df.shape[0]} days)")

    if enable_save_plot:
        file_name = get_file_name('scatter_matrix', name=name)
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    if enable_show_plot:
        plt.show()

    plt.cla()
    plt.clf()

    return file_path


def hist_norm(data, bins=10, enable_show_plot=True, enable_save_plot=False, name=None):
    """hist 分布图及正太分布曲线"""
    n, bins_v, file_name = None, None, None

    # """绘图函数"""
    # ax = pct_change_s.hist(bins=50, density=1)
    fig, ax = plt.subplots()
    # the histogram of the data
    _, _, patches = ax.hist(data, bins, density=True)
    n, bins_v, mean, std = plot_norm(data, bins=bins, ax=ax)
    # ax.set_xlabel('pct change')
    # ax.set_ylabel('change rate')
    ax.set_title(f"{'Data' if name is None else name} Histogram (mean={mean:.4f} std={std:.4f})")

    if enable_save_plot:
        file_name = get_file_name(f'hist', name=name)
        rr_plot_file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(rr_plot_file_path, dpi=75)

    if enable_show_plot:
        plt.show()

    plt.cla()
    plt.clf()

    return n, bins_v, rr_plot_file_path


def plot_norm(data: pd.Series, bins=10, ax=None, is_show_plot=None):
    """
    显示当前数据的正太分布曲线
    :param data:
    :param bins: bar 数量
    :param ax: 如果为None，则新建一个画布
    :param is_show_plot: 是否展示
    :return: n, bins_v, mean, std
    """
    if ax is None:
        fig, ax = plt.subplots()
        if is_show_plot is None:
            is_show_plot = True

    if is_show_plot is None:
        is_show_plot = False

    n, bins_v = np.histogram(data, bins=bins)

    mu = data.mean()  # mean of distribution
    sigma = data.std()  # standard deviation of distribution
    # def norm_func(x, mu, sigma):
    #     pdf = np.exp(-((x - mu)**2)/(2*sigma**2)) / (sigma * np.sqrt(2*np.pi))
    #     return pdf
    # y = norm_func(bins, mu, sigma)  # 与 mlab.normpdf(bins, mu, sigma) 相同
    # y = mlab.normpdf(bins, mu, sigma)
    y = stats.norm.pdf(bins_v, loc=mu, scale=sigma)
    ax.plot(bins_v, y, '--')
    if is_show_plot:
        plt.show()

    return n, bins_v, mu, sigma


def hist_norm_sns(data, bins=10):
    """sns 方式 hist 分布图及正太分布曲线"""
    sns.set_palette("hls")
    sns.distplot(
        data, bins=bins,
        fit=stats.norm,
        kde_kws={"color": "darkorange", "lw": 1, "label": "KDE", "linestyle": "--"},
        hist_kws={"color": "darkblue"})
    plt.show()


def wave_hist(df: pd.DataFrame, columns=None, bins=50, figure_4_each_col=True,
              col_transfer_dic: (dict, None) = None, enable_show_plot=True, enable_save_plot=False, name=None):
    """
    波动率分布图
    :param df:
    :param perf_stats: df.calc_stats()
    :param columns: 显示哪些列
    :param bins: bar 数量
    :param figure_4_each_col: 每一个col单独一张图
    :param col_transfer_dic: 列值转换
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if columns is not None:
        data_df = df.loc[:, [_ for _ in columns if _ in df.columns]].copy()
    else:
        data_df = df.copy()

    data_df.dropna(inplace=True)
    if col_transfer_dic is not None:
        for func, col_name_list in col_transfer_dic.items():
            if func == 'return':
                rename_dic = {}
                for col_name in col_transfer_dic[func]:
                    if col_name not in data_df:
                        continue
                    rename_dic[col_name] = f"{col_name} {func}"
                    data_df.loc[:, col_name] = data_df[col_name].to_returns()
                else:
                    data_df.rename(columns=rename_dic, inplace=True)

            elif func == 'pct_change':
                rename_dic = {}
                for col_name in col_transfer_dic[func]:
                    if col_name not in data_df:
                        continue
                    rename_dic[col_name] = f"{col_name} {func}"
                    data_df.loc[:, col_name] = data_df[col_name].pct_change()
                else:
                    data_df.rename(columns=rename_dic, inplace=True)
            else:
                rename_dic = {}
                for col_name in col_transfer_dic[func]:
                    if col_name not in data_df:
                        continue
                    rename_dic[col_name] = f"{col_name} {str(func)}"
                    data_df.loc[:, col_name] = func(data_df[col_name])
                else:
                    data_df.rename(columns=rename_dic, inplace=True)

        data_df.dropna(inplace=True)

    columns = list(data_df.columns)
    n_bins_dic, file_path = {}, None
    if figure_4_each_col:
        for col_name in columns:
            data = data_df[col_name]
            try:
                data = data[~np.isinf(data)].dropna()
            except TypeError as exp:
                logger.exception('column %s 数据类型无效，无法进行 np.isinf 计算', col_name)
                raise exp from exp

            n, bins_v, file_path_sub = hist_norm(
                data, bins=bins, enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot, name=col_name)
            n_bins_dic[col_name] = (n, bins_v)
            if file_path is None:
                file_path = [file_path_sub]
            else:
                file_path.append(file_path_sub)

    else:

        ax = data_df.hist(bins=bins)
        if isinstance(ax, np.ndarray):
            # ax 总是一个 偶数 出现，因此，需要进行一次长度对齐
            ax_list = list(itertools.chain(*ax))[:len(columns)]
        else:
            ax_list = [ax]

        for col_name, ax_sub in zip(columns, ax_list):
            data = data_df[col_name]
            try:
                data = data[~np.isinf(data)].dropna()
            except TypeError as exp:
                logger.exception('column %s 数据类型无效，无法进行 np.isinf 计算', col_name)
                raise exp from exp

            n, bins_v, mean, std = plot_norm(data, bins=bins, ax=ax_sub)
            # 字太多无法显示
            # ax_sub.set_title(f"{col_name}\n(mean={mean:.4f} std={std:.4f})")
            n_bins_dic[col_name] = (n, bins_v)

        if enable_save_plot:
            file_name = get_file_name(f'hist', name=name)
            file_path = os.path.join(get_cache_folder_path(), file_name)
            plt.savefig(file_path, dpi=75)

        if enable_show_plot:
            plt.show()

        plt.cla()
        plt.clf()

    return n_bins_dic, file_path


def _test_wave_hist():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    col_transfer_dic = {
        'return': ['open', 'high', 'low', 'close', 'volume']
    }
    n_bins_dic, file_path = wave_hist(df, figure_4_each_col=False, col_transfer_dic=col_transfer_dic)


def hist_n_rr(df: pd.DataFrame, n_days, columns=None, bins=50,
              enable_show_plot=True, enable_save_plot=False, name=None):
    """
    未来N日收益率分布图
    :param df:
    :param n_days: 计算未来 N 日的收益率最高值，最低值
    :param columns: 显示哪些列
    :param bins: bar 数量
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if columns is not None:
        data_df = df.loc[:, columns].dropna().copy()
    else:
        data_df = df.dropna().copy()

    column_name_list, df_dic, quantile_dic, n_bins_dic, file_path_dic = list(data_df.columns), {}, {}, {}, {}
    for n_day in n_days:
        # 计算各个列未来N日收益率波动
        max_df = data_df.rename(
            columns={_: _ + ' max rr' for _ in column_name_list}
        ).rolling(n_day).apply(
            lambda x: max(x / x[0]) - 1, raw=True).shift(-(n_day - 1))
        min_df = data_df.rename(
            columns={_: _ + ' min rr' for _ in column_name_list}
        ).rolling(n_day).apply(
            lambda x: min(x / x[0]) - 1, raw=True).shift(-(n_day - 1))
        # 合并数据
        merged_df = max_df.join(min_df).dropna()
        for col_name in column_name_list:
            col_names = [col_name + ' max rr', col_name + ' min rr']
            new_df = merged_df[col_names].copy()
            df_list = [new_df]
            # 计算收益率分位数信息
            idxmax_s = abs(new_df).idxmax(axis=1)
            # df.loc[idxmax_s != 'close max rr', 'close max rr'] = np.nan
            # df.loc[idxmax_s != 'close min rr', 'close min rr'] = np.nan
            label_quantile_dic = {}
            for label in new_df.columns:
                data_s = new_df.loc[idxmax_s == label, label].dropna()
                df_list.append(data_s)
                label_quantile_dic[label] = data_s.quantile([0.2, 0.33, 0.5, 0.66, 0.8])

            df_dic[(n_day, col_name)] = df_list
            quantile_dic[(n_day, col_name)] = pd.DataFrame(label_quantile_dic).T

    # 输出图片
    for (n_day, col_name), df_list in df_dic.items():
        hist_kwargs = dict(histtype='stepfilled', alpha=0.8, bins=bins)
        n_list, bins_v_list, axis_list, labels = [], [], [], []
        for data_s in df_list[1:]:
            n, bins_v, axis = plt.hist(data_s.dropna(), **hist_kwargs)
            labels.append(data_s.name)
            axis_list.extend(axis)

        plt.legend(axis_list, labels, loc='upper right')
        if name is None:
            title = f'hist of max/min {col_name} rr in {n_day} days'
        else:
            title = f'{name} hist of max/min {col_name} rr in {n_day} days'

        plt.suptitle(title)
        n_bins_dic[(n_day, col_name)] = (n_list, bins_v_list)

        if enable_save_plot:
            file_name = get_file_name(f'hist {n_day} {col_name} future', name=name)
            file_path = os.path.join(get_cache_folder_path(), file_name)
            plt.savefig(file_path, dpi=75)
            file_path_dic[(n_day, col_name)] = file_path

        if enable_show_plot:
            plt.show()

        plt.cla()
        plt.clf()

    ret_dic = dict(df_dic=df_dic, n_bins_dic=n_bins_dic, quantile_dic=quantile_dic)
    return ret_dic, file_path_dic


def _test_hist_n_rr():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    ret_dic, file_path_dic = hist_n_rr(df, n_days=[3, 5], columns=['close'], enable_save_plot=True)
    for k, v in ret_dic['quantile_dic'].items():
        logger.info('%s -> \n%s', k, v)


def label_distribution(close_df: pd.DataFrame, min_rr: float, max_rr: float, max_future: int,
                       name=None, **enable_kwargs):
    """
    输出分类标签在行情图中的分布情况
    :param close_df:
    :param min_rr:
    :param max_rr:
    :param max_future:为空则进行2分类，不为空则3分类
    :param enable_kwargs:
    :param name:
    :return:
    """
    logger.debug('%s [%f ~ %f] max_future=%d, name="%s"', close_df.shape, min_rr, max_rr, max_future, name)
    value_arr = close_df.to_numpy()
    if max_future is None:
        target_arr = calc_label2(value_arr, min_rr, max_rr, one_hot=False, dtype='int')
    else:
        target_arr = calc_label3(value_arr, min_rr, max_rr, max_future=max_future, one_hot=False, dtype='int')

    ax = close_df.plot()
    plt.suptitle(f'label [{min_rr * 100:.2f}% ~ {max_rr * 100:.2f}%]')
    x_values = list(close_df.index)
    colors = [None, '#2ca02c', '#d62728']

    value_count = len(value_arr)
    label_list = list(set(target_arr))
    distribution_rate_df = pd.DataFrame(
        {_: {'pct': sum(target_arr == _) / value_count} for _ in label_list}, columns=label_list)
    for label, range_from, range_to in get_range_num_iter(target_arr):
        color = colors[int(label)]
        if color is None:
            continue
        p = plt.axvspan(x_values[range_from], x_values[range_to], facecolor=color, alpha=0.5)

    file_name = get_file_name(f'label distribution', name=name)
    file_path = _plot_or_show(file_name=file_name, **enable_kwargs)
    return distribution_rate_df, file_path


def _plot_or_show(enable_save_plot=True, enable_show_plot=True, file_name=None):
    if enable_save_plot:
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    if enable_show_plot:
        plt.show()

    plt.cla()
    plt.clf()
    return file_path


def _test_label_distribution():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    distribution_rate_df, file_path = label_distribution(df['close'], min_rr=-0.01, max_rr=0.01, max_future=3)
    logger.info('\n%s', distribution_rate_df)
    logger.info(file_path)


def get_range_num_iter(arr):
    val_cur, range_from, range_to = arr[0], 0, None
    for range_to, val in enumerate(arr[1:], start=1):
        if val != val_cur:
            yield val_cur, range_from, range_to
            range_from = range_to
            val_cur = val
    else:
        range_to += 1
        yield val_cur, range_from, range_to


def _test_n_days_rr_distribution():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    ret_dic, file_path_dic, enable_kwargs = {}, {}, dict(enable_save_plot=True, enable_show_plot=True)
    tmp_dic, file_path = hist_n_rr(df, n_days=[3, 5], columns=['close'], **enable_kwargs)
    ret_dic['hist_future_n_rr'] = tmp_dic
    file_path_dic['hist_future_n_rr'] = file_path

    from collections import defaultdict
    file_path_dic['label_distribution'] = defaultdict(dict)
    ret_dic['label_distribution'] = defaultdict(dict)
    for (n_day, col_name), quantile_df in tmp_dic['quantile_dic'].items():
        path_dic = file_path_dic['label_distribution'][(n_day, col_name)]
        distribution_dic = ret_dic['label_distribution'][(n_day, col_name)]
        col_count = quantile_df.shape[1]
        for n in range(col_count):
            max_rr = quantile_df.iloc[0, n]
            min_rr = quantile_df.iloc[1, col_count - n - 1]
            distribution_rate_df, file_path = label_distribution(
                df['close'], min_rr=min_rr, max_rr=max_rr, max_future=n_day,
                name=f"{col_name}[{min_rr*100:.2f}%-{max_rr*100:.2f}%]", **enable_kwargs)
            path_dic[(min_rr, max_rr)] = file_path
            distribution_dic[(min_rr, max_rr)] = distribution_rate_df


if __name__ == "__main__":
    # _test_wave_hist()
    # _test_hist_n_rr()
    # _test_label_distribution()
    _test_n_days_rr_distribution()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/4/7 16:25
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import pandas as pd
from functools import lru_cache

pd.set_option('display.width', 240)
pd.set_option('display.max_columns', 20)
pd.set_option('display.float_format', '{:,.4f}'.format)


@lru_cache(maxsize=1)
def get_report_folder_path() -> str:
    import os
    from ibats_common import module_root_path
    _report_file_path = os.path.join(module_root_path, 'analysis', 'report')
    if not os.path.exists(_report_file_path):
        os.makedirs(_report_file_path)

    return _report_file_path


@lru_cache(maxsize=1)
def get_cache_folder_path() -> str:
    import os
    from ibats_common import module_root_path
    _report_file_path = os.path.join(module_root_path, 'analysis', 'report', "_cache_")
    if not os.path.exists(_report_file_path):
        os.makedirs(_report_file_path)

    return _report_file_path




#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/3/8 12:23
@File    : orm_stg_run_status_detail_test.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_common.backend.orm import *
import unittest
from ibats_common.test.orm_init_test import InitTest


class PosStatusDetailTest(unittest.TestCase):  # 继承unittest.TestCase
    def tearDown(self):
        pass

    def setUp(self):
        # 每个测试用例执行之前做操作
        pass

    @classmethod
    def tearDownClass(cls):
        # 必须使用 @ classmethod装饰器, 所有test运行完后运行一次
        pass

    @classmethod
    def setUpClass(cls):
        # from ibats_common.config import update_db_config, ConfigBase
        # update_db_config({
        #     ConfigBase.DB_SCHEMA_IBATS: 'mysql://mg:Dcba1234@localhost/' + ConfigBase.DB_SCHEMA_IBATS,
        # })
        # 必须使用@classmethod 装饰器,所有test运行前运行一次
        init()
        global engine_ibats
        engine_ibats = engines.engine_ibats

    def test_func(self):
        pass

    def test_create_by_trade_detail(self):
        trade = PosStatusDetailTest.add_trade()
        status = PosStatusDetail.create_by_trade_detail(trade)
        self.assertIsInstance(status, PosStatusDetail)
        self.assertEqual(status.direction, trade.direction)
        self.assertEqual(status.symbol, trade.symbol)
        self.assertEqual(status.position, trade.trade_vol)
        self.assertEqual(status.position_chg, trade.trade_vol)
        self.assertEqual(status.margin, trade.margin)
        self.assertEqual(status.margin_chg, trade.margin)
        self.assertEqual(status.floating_pl // 1, -trade.commission // 1)
        self.assertEqual(status.floating_pl_rate, -trade.commission / status.margin)  # 初次建仓时浮动收益就等于手续费 / 保证金
        self.assertEqual(status.floating_pl_chg, status.floating_pl)
        self.assertEqual(status.floating_pl_cum, status.floating_pl)
        self.assertEqual(status.cashflow, status.cashflow)
        self.assertEqual(status.cashflow_cum, status.cashflow_cum)
        self.assertEqual(status.calc_mode, status.calc_mode)
        self.assertEqual(status.rr, status.floating_pl_rate)
        self.assertEqual(status.commission, trade.commission)
        self.assertEqual(status.position_date_type, PositionDateType.Today.value)
        self.assertEqual(status.multiple, trade.multiple)
        self.assertEqual(status.margin_ratio, trade.margin_ratio)

    def test_create_self(self):
        trade = PosStatusDetailTest.add_trade()
        status = PosStatusDetail.create_by_trade_detail(trade)
        status2 = status.create_by_self()
        self.assertIsInstance(status2, PosStatusDetail)
        self.assertEqual(status2.direction, status.direction)
        self.assertEqual(status2.symbol, status.symbol)
        self.assertEqual(status2.position, status.position)
        self.assertEqual(status2.position_chg, 0)
        self.assertEqual(status2.margin, status.margin)
        self.assertEqual(status2.margin_chg, 0)
        self.assertEqual(status2.floating_pl, status.floating_pl)
        self.assertEqual(status2.floating_pl_rate, 0)
        self.assertEqual(status2.floating_pl_chg, 0)
        self.assertEqual(status2.floating_pl_cum, status.floating_pl_cum)
        self.assertEqual(status2.cashflow, 0)
        self.assertEqual(status2.cashflow_cum, status.cashflow_cum)
        self.assertEqual(status2.calc_mode, status.calc_mode)
        self.assertEqual(status2.rr, 0)
        self.assertEqual(status2.commission, 0)
        self.assertEqual(status2.position_date_type, status.position_date_type)
        self.assertEqual(status2.multiple, status.multiple)
        self.assertEqual(status2.margin_ratio, status.margin_ratio)

    def test_update_by_trade_detail1(self):
        # 多头加仓
        trade = PosStatusDetailTest.add_trade()
        status = PosStatusDetail.create_by_trade_detail(trade)
        trade2 = PosStatusDetailTest.add_trade_4_test_update_by_trade_detail1()
        self.assertEqual(trade.trade_vol * 2, trade2.trade_vol)
        status2 = status.update_by_trade_detail(trade2)
        self.assertIsInstance(status2, PosStatusDetail)
        self.assertEqual(status2.direction, trade.direction)
        self.assertEqual(status2.symbol, trade.symbol)
        self.assertEqual(status2.position, trade.trade_vol + trade2.trade_vol)
        self.assertEqual(status2.position_chg, trade2.trade_vol)
        self.assertEqual(status2.margin, trade.margin + trade2.margin)
        self.assertEqual(status2.margin_chg, trade2.margin)
        self.assertEqual(status2.floating_pl // 1, -(trade.commission + trade2.commission) // 1)
        # 初次建仓时浮动收益就等于手续费 / 保证金 "*10000//1" 指在 万分之一以上精度相等即可
        self.assertEqual(status2.floating_pl_rate * 10000 // 1, -(
                    trade.commission + trade2.commission) / status2.margin * 10000 // 1)
        self.assertEqual(status2.floating_pl_chg, status2.floating_pl - status.floating_pl)
        self.assertEqual(status2.floating_pl_cum, status2.floating_pl)
        self.assertEqual(status2.cashflow, - status2.margin_chg - status2.commission)
        self.assertEqual(status2.cashflow_cum, status.cashflow_cum + status2.cashflow)
        self.assertEqual(status2.calc_mode, status.calc_mode)
        self.assertEqual(status2.rr * 10000 // 1, status2.floating_pl_rate * 10000 // 1)
        self.assertEqual(status2.commission, trade2.commission)
        self.assertEqual(status2.commission_tot, trade.commission + trade2.commission)
        self.assertEqual(status2.position_date_type, PositionDateType.Today.value)
        self.assertEqual(status2.multiple, trade.multiple)
        self.assertEqual(status2.margin_ratio, trade.margin_ratio)

    def test_update_by_trade_detail2(self):
        # 多头减仓
        trade = PosStatusDetailTest.add_trade()
        status = PosStatusDetail.create_by_trade_detail(trade)
        trade2 = PosStatusDetailTest.add_trade_4_test_update_by_trade_detail2()
        self.assertEqual(trade2.trade_vol, trade.trade_vol / 2)
        status2 = status.update_by_trade_detail(trade2)
        self.assertIsInstance(status2, PosStatusDetail)
        self.assertEqual(status2.direction, trade.direction)
        self.assertEqual(status2.symbol, trade.symbol)
        self.assertEqual(status2.position, trade.trade_vol - trade2.trade_vol)
        self.assertEqual(status2.position_chg, -trade2.trade_vol)
        self.assertEqual(status2.margin, trade.margin - trade2.margin)
        self.assertEqual(status2.margin_chg, -trade2.margin)
        self.assertEqual(status2.floating_pl // 1, -(trade.commission + trade2.commission) // 1)
        # 初次建仓时浮动收益就等于手续费 / 保证金 "*10000//1" 指在 万分之一以上精度相等即可
        self.assertEqual(status2.floating_pl_rate * 10000 // 1, -(
                    trade.commission + trade2.commission) / status2.margin * 10000 // 1)
        self.assertEqual(status2.floating_pl_chg, status2.floating_pl - status.floating_pl)
        self.assertEqual(status2.floating_pl_cum, status2.floating_pl)
        self.assertEqual(status2.cashflow, - status2.margin_chg - status2.commission)
        self.assertEqual(status2.cashflow_cum, status.cashflow_cum + status2.cashflow)
        self.assertEqual(status2.calc_mode, status.calc_mode)
        self.assertEqual(status2.rr * 10000 // 1, status2.floating_pl_rate * 10000 // 1)
        self.assertEqual(status2.commission, trade2.commission)
        self.assertEqual(status2.commission_tot, trade.commission + trade2.commission)
        self.assertEqual(status2.position_date_type, PositionDateType.Today.value)
        self.assertEqual(status2.multiple, trade.multiple)
        self.assertEqual(status2.margin_ratio, trade.margin_ratio)

    def test_update_by_trade_detail3(self):
        # 多头平仓
        trade = PosStatusDetailTest.add_trade()
        status = PosStatusDetail.create_by_trade_detail(trade)
        trade2 = PosStatusDetailTest.add_trade_4_test_update_by_trade_detail3()
        self.assertEqual(trade2.trade_vol, trade.trade_vol)
        status2 = status.update_by_trade_detail(trade2)
        self.assertIsInstance(status2, PosStatusDetail)
        self.assertEqual(status2.direction, trade.direction)
        self.assertEqual(status2.symbol, trade.symbol)
        self.assertEqual(status2.position, 0)
        self.assertEqual(status2.position_chg, -trade2.trade_vol)
        self.assertEqual(status2.margin, 0)
        self.assertEqual(status2.margin_chg, -trade2.margin)
        self.assertEqual(status2.floating_pl // 1, -(trade.commission + trade2.commission) // 1)
        # 初次建仓时浮动收益就等于手续费 / 保证金 "*10000//1" 指在 万分之一以上精度相等即可
        self.assertEqual(status2.floating_pl_rate * 10000 // 1, -(
                    trade.commission + trade2.commission) / status.margin * 10000 // 1)
        self.assertEqual(status2.floating_pl_chg, status2.floating_pl - status.floating_pl)
        self.assertEqual(status2.floating_pl_cum, status2.floating_pl)
        self.assertEqual(status2.cashflow, - status2.margin_chg - status2.commission)
        self.assertEqual(status2.cashflow_cum, status.cashflow_cum + status2.cashflow)
        self.assertEqual(status2.calc_mode, status.calc_mode)
        self.assertEqual(status2.rr * 10000 // 1, status2.floating_pl_rate * 10000 // 1)
        self.assertEqual(status2.commission, trade2.commission)
        self.assertEqual(status2.commission_tot, trade.commission + trade2.commission)
        self.assertEqual(status2.position_date_type, PositionDateType.Today.value)
        self.assertEqual(status2.multiple, trade.multiple)
        self.assertEqual(status2.margin_ratio, trade.margin_ratio)

    @staticmethod
    def add_trade(stg_run_id=None):
        order = InitTest.add_order(stg_run_id)
        trade = TradeDetail.create_by_order_detail(order)
        return trade

    @staticmethod
    def add_trade_4_test_update_by_trade_detail1():
        order = InitTest.add_order()
        order.order_vol += order.order_vol
        trade = TradeDetail.create_by_order_detail(order)
        trade.trade_date += timedelta(days=1)
        trade.trade_dt += timedelta(days=1)
        return trade

    @staticmethod
    def add_trade_4_test_update_by_trade_detail2():
        order = InitTest.add_order()
        order.order_vol = order.order_vol / 2
        order.action = Action.Close.value
        trade = TradeDetail.create_by_order_detail(order)
        trade.trade_date += timedelta(days=1)
        trade.trade_dt += timedelta(days=1)
        return trade

    @staticmethod
    def add_trade_4_test_update_by_trade_detail3():
        order = InitTest.add_order()
        order.action = Action.Close.value
        trade = TradeDetail.create_by_order_detail(order)
        trade.trade_date += timedelta(days=1)
        trade.trade_dt += timedelta(days=1)
        return trade


if __name__ == '__main__':
    unittest.main()  # 运行所有的测试用例



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-5-24 下午3:16
@File    : label.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import numpy as np
from keras.utils import to_categorical


def calc_label2(value_arr: np.ndarray, min_rr: float, max_rr: float, one_hot=True, dtype='float32'):
    """
    根据时间序列数据 pct_arr 计算每一个时点目标标示 0 1 2
    计算方式：
    当某一点未来波动首先 < 下届 min_pct，则标记为： [0, 1, 0]
    当某一点未来波动首先 > 上届 max_pct，则标记为： [0, 0, 1]
    :param value_arr:
    :param min_rr:
    :param max_rr:
    :param one_hot:
    :param dtype:
    :return:
    """

    value_arr[np.isnan(value_arr)] = 0
    arr_len = value_arr.shape[0]
    # target_arr = np.zeros((arr_len, 2))
    label_arr = np.zeros(arr_len, dtype=dtype)
    for i in range(arr_len):
        base = value_arr[i]
        for j in range(i + 1, arr_len):
            result = value_arr[j] / base - 1
            if result < min_rr:
                label_arr[i] = 1
                break
            elif result > max_rr:
                label_arr[i] = 2
                break
    if one_hot:
        # target_arr = pd.get_dummies(label_arr)
        target_arr = to_categorical(label_arr, num_classes=3, dtype=dtype)
    else:
        target_arr = label_arr

    return target_arr


def _test_calc_label2(show_plt=True):
    import matplotlib.pyplot as plt
    i_s = np.arange(0, 20, 0.1)
    price_arr = np.cos(i_s) + 3  # cos(x) + 5
    one_hot = False
    labels = calc_label2(price_arr, -0.1, 0.1, one_hot=one_hot)
    if show_plt:
        plt.plot(i_s, labels, 'r',
                 i_s, price_arr, 'b--',
                 )
        plt.show()

    one_hot = True
    labels = calc_label2(price_arr, -0.1, 0.1, one_hot=one_hot)
    if show_plt:
        plt.plot(i_s, np.argmax(labels, axis=1), 'r',
                 i_s, price_arr, 'b--',
                 )
        plt.show()


def calc_label3(value_arr: np.ndarray, min_rr: float, max_rr: float, max_future: int, one_hot=True, dtype='float32'):
    """
    根据时间序列数据 pct_arr 计算每一个时点目标标示 -1 0 1
    计算方式：
    max_future 内未突破min_pct 或 max_pct，则标记为:  [1, 0, 0]
    当某一点未来波动首先 < 下届 min_pct，则标记为:      [0, 1, 0]
    当某一点未来波动首先 > 上届 max_pct，则标记为:      [0, 0, 1]
    :param value_arr:
    :param min_rr:
    :param max_rr:
    :param max_future:
    :param one_hot:
    :param dtype:
    :return:
    """

    value_arr[np.isnan(value_arr)] = 0
    arr_len = value_arr.shape[0]
    # target_arr = np.zeros((arr_len, 2))
    label_arr = np.zeros(arr_len, dtype=dtype)
    for i in range(arr_len):
        base = value_arr[i]
        for j in range(i + 1, arr_len):
            result = value_arr[j] / base - 1
            if result < min_rr:
                label_arr[i] = 1
                break
            elif result > max_rr:
                label_arr[i] = 2
                break
            elif max_future is not None and j-i >= max_future:
                # 超过最大检测日期，则被认为是震荡势
                # label_arr[i] = 0
                break
    if one_hot:
        # target_arr = pd.get_dummies(label_arr)
        target_arr = to_categorical(label_arr, num_classes=3, dtype=dtype)
    else:
        target_arr = label_arr

    return target_arr


def _test_calc_label3(show_plt=True):
    import matplotlib.pyplot as plt
    i_s = np.arange(0, 20, 0.1)
    price_arr = np.cos(i_s) + 3  # cos(x) + 5
    one_hot = False
    labels = calc_label3(price_arr, -0.1, 0.1, max_future=5, one_hot=one_hot)
    if show_plt:
        plt.plot(i_s, labels, 'r',
                 i_s, price_arr, 'b--',
                 )
        plt.show()

    one_hot = True
    labels = calc_label3(price_arr, -0.1, 0.1, max_future=5, one_hot=one_hot)
    if show_plt:
        plt.plot(i_s, np.argmax(labels, axis=1), 'r',
                 i_s, price_arr, 'b--',
                 )
        plt.show()


if __name__ == "__main__":
    _test_calc_label2()
    # _test_calc_label3()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/4/14 12:23
@File    : plot2_test.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import unittest
from ibats_common.analysis.plot_db import show_order, show_trade


class SomeTest(unittest.TestCase):  # 继承unittest.TestCase
    def tearDown(self):
        pass

    def setUp(self):
        # 每个测试用例执行之前做操作
        pass

    @classmethod
    def tearDownClass(cls):
        # 必须使用 @ classmethod装饰器, 所有test运行完后运行一次
        pass

    @classmethod
    def setUpClass(cls):
        print('set up')
        from ibats_common.example.ma_cross_stg import _test_use
        cls.stg_run_id = _test_use()

    def test_func(self):
        self.assertEqual(1, 1)

    def test_order_func(self):
        data_dict = show_order(stg_run_id=self.stg_run_id)
        self.assertEqual(len(data_dict), 1)
        key = list(data_dict.keys())[0]
        self.assertEqual(len(data_dict[key]['md']), 1)
        self.assertEqual(len(data_dict[key]['long_open_or_short_close']), 1)
        self.assertEqual(len(data_dict[key]['short_open_or_long_close']), 1)
        self.assertGreater(len(data_dict[key]['buy_sell_point_pair']), 2)
        self.assertGreater(len(data_dict[key]['sell_buy_point_pair']), 2)

    def test_trade_func(self):
        data_dict = show_trade(stg_run_id=self.stg_run_id)
        self.assertEqual(len(data_dict), 1)
        key = list(data_dict.keys())[0]
        self.assertEqual(len(data_dict[key]['md']), 1)
        self.assertEqual(len(data_dict[key]['long_open_or_short_close']), 1)
        self.assertEqual(len(data_dict[key]['short_open_or_long_close']), 1)
        self.assertGreater(len(data_dict[key]['buy_sell_point_pair']), 2)
        self.assertGreater(len(data_dict[key]['sell_buy_point_pair']), 2)


if __name__ == '__main__':
    unittest.main()  # 运行所有的测试用例



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-4-19 上午8:40
@File    : ai_stg.py
@contact : mmmaaaggg@163.com
@desc    : 简单的 RNN LSTM 构建策略模型，运行该模型需要首先安装 TensorFlow 包
pip3 install tensorflow sklearn tflearn
"""
import os
import random
from datetime import datetime

import numpy as np
import tensorflow as tf
import tflearn
from ibats_utils.mess import get_last_idx
from sklearn.model_selection import train_test_split
from tflearn import conv_2d, max_pool_2d, local_response_normalization, fully_connected

from ibats_common import module_root_path
from ibats_common.backend.factor import get_factor
from ibats_common.backend.label import calc_label2
from ibats_common.common import BacktestTradeMode, ContextKey, Direction, CalcMode
from ibats_common.example.data import get_trade_date_series, get_delivery_date_series
from ibats_common.strategy import StgBase
from ibats_common.strategy_handler import strategy_handler_factory
from ibats_local_trader.agent.md_agent import *
from ibats_local_trader.agent.td_agent import *

logger = logging.getLogger(__name__)


class AIStg(StgBase):

    def __init__(self, instrument_type, unit=1):
        super().__init__()
        self.unit = unit
        self.input_size = 39
        self.batch_size = 64
        self.n_step = 20
        self.output_size = 2
        self.n_hidden_units = 10
        self.lr = 0.006
        self._model = None
        self._session = None
        self.train_validation_rate = 0.8
        self.enable_load_model_if_exist = False
        self.n_epoch = 20
        self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = None, None, None, None
        self.classify_wave_rate = 0.0025
        self.predict_test_random_state = None
        datetime_str = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
        self.base_folder_path = folder_path = os.path.join(module_root_path, f'tf_saves_{datetime_str}')
        model_folder_path = os.path.join(folder_path, 'model_tfls')
        if not os.path.exists(model_folder_path):
            os.makedirs(model_folder_path, exist_ok=True)
        file_path = os.path.join(
            model_folder_path,
            f"model_{int(self.classify_wave_rate * 10000)}.tfl")
        self.model_file_path = file_path
        self.checkpoint_path = os.path.join(
            model_folder_path,
            f'model_{int(self.classify_wave_rate * 10000)}.tfl.ckpt')
        tensorboard_dir = os.path.join(folder_path, f'tensorboard_logs')
        if not os.path.exists(tensorboard_dir):
            os.makedirs(tensorboard_dir)
        self.tensorboard_dir = tensorboard_dir
        self.trade_date_last_train = None
        self.retrain_period = 10
        # 其他辅助信息
        self.trade_date_series = get_trade_date_series()
        self.delivery_date_series = get_delivery_date_series(instrument_type)

    @property
    def session(self):
        return self.get_session()

    def get_session(self, renew=False, as_default=True, close_last_one_if_renew=True):
        if renew or self._session is None:
            if self.model is None:
                raise ValueError('model 需要先于 session 被创建')
            if close_last_one_if_renew and self._session is not None:
                self._session.close()
            self._session = tf.Session()
            self._session.run(tf.global_variables_initializer())
            # if as_default:
            #     self._session.as_default()
            #     logger.info('default session:%s', self._session.graph)
        return self._session

    @property
    def model(self) -> tflearn.models.DNN:
        return self.get_model()

    def get_model(self, rebuild_model=False) -> tflearn.models.DNN:
        if self._model is None or rebuild_model:
            tf.reset_default_graph()
            self._model = self._build_model()
        return self._model

    def get_factor_array(self, md_df: pd.DataFrame, tail_n=None):
        if tail_n is not None:
            md_df = md_df.tail(tail_n)
        df = md_df[~md_df['close'].isnull()]
        # df = md_df[~md_df['close'].isnull()][[
        #     'open', 'high', 'low', 'close', 'volume', 'oi', 'warehousewarrant', 'termstructure']]
        # df['ma5'] = df['close'].rolling(window=5).mean()
        # df['ma10'] = df['close'].rolling(window=10).mean()
        # df['ma20'] = df['close'].rolling(window=20).mean()
        # df['pct_change_vol'] = df['volume'].pct_change()
        # df['pct_change'] = df['close'].pct_change()

        factors = df.fillna(0).to_numpy()
        if self.input_size is None or self.input_size != factors.shape[1]:
            self.input_size = factors.shape[1]
            self.n_hidden_units = self.input_size * 2
            logger.info("set input_size: %d", self.input_size)
            logger.info("set n_hidden_units: %d", self.n_hidden_units)

        return factors

    def get_x_y(self, md_df):
        factors = self.get_factor_array(md_df)
        price_arr = factors[:, 0]
        self.input_size = factors.shape[1]
        # ys_all = self.calc_y_against_future_data(price_arr, -0.01, 0.01)
        # ys_all = self.calc_y_against_future_data(price_arr, -self.classify_wave_rate, self.classify_wave_rate)
        ys_all = calc_label2(price_arr, -self.classify_wave_rate, self.classify_wave_rate, one_hot=True)
        if ys_all.shape[1] == 3:
            idx_last_available_label = get_last_idx(ys_all, lambda x: x[1] == 1 or x[2] == 1)
            if idx_last_available_label is not None:
                factors = factors[:idx_last_available_label + 1, :]
                ys_all = ys_all[:idx_last_available_label + 1, [1, 2]]

                range_from = self.n_step
                range_to = idx_last_available_label + 1
            else:
                range_from = self.n_step
                range_to = factors.shape[0]
        else:
            range_from = self.n_step
            range_to = factors.shape[0]

        xs = np.zeros((range_to - range_from, self.n_step, self.input_size, 1))
        for num, index in enumerate(range(range_from, range_to)):
            xs[num, :, :, :] = factors[(index - self.n_step):index, :, np.newaxis]

        ys = ys_all[range_from:range_to, :]

        return xs, ys

    def get_batch_xs(self, factors: np.ndarray, index=None):
        """
        取 batch_xs
        :param factors:
        :param index: 样本起始坐标，如果为None，则默认取尾部一组样本
        :return:
        """
        if index is None:
            index = factors.shape[0] - 1

        batch_xs = np.zeros((1, self.n_step, self.input_size, 1))
        batch_xs[0, :, :, :] = factors[(index - self.n_step + 1):(index + 1), :, np.newaxis]

        return batch_xs

    def get_batch_by_random(self, factors: np.ndarray, labels: np.ndarray):
        """
        够在一系列输入输出数据集
        xs： 两条同频率，不同位移的sin曲线
        ys_value： 目标是一条cos曲线
        ys: ys_value 未来涨跌标识
        i_s：X 序列
        """
        xs = np.zeros((self.batch_size, self.n_step, self.input_size))
        ys = np.zeros((self.batch_size, self.output_size))
        # available_batch_size, num = 0, 0
        samples_index = random.sample(range(self.n_step - 1, factors.shape[0] - 1), self.batch_size)
        examples_index_list = []
        for available_batch_size, index in enumerate(samples_index):
            tmp = factors[(index - self.n_step + 1):(index + 1), :]
            if tmp.shape[0] < self.n_step:
                break
            xs[available_batch_size, :, :] = tmp
            ys[available_batch_size, :] = labels[index, :]
            examples_index_list.append(index)
            if available_batch_size + 1 >= self.batch_size:
                available_batch_size += 1
                break

        # returned xs, ys_value and shape (batch, step, input)
        return xs, ys, examples_index_list

    def _build_model(self) -> tflearn.models.DNN:
        # hyperparameters
        # lr = LR
        # batch_size = BATCH_SIZE
        #
        # n_inputs = INPUT_SIZE  # MNIST data input (img shape 28*28)
        # n_step = TIME_STEPS  # time steps
        # n_hidden_units = CELL_SIZE  # neurons in hidden layer
        # n_classes = OUTPUT_SIZE  # MNIST classes (0-9 digits)
        # Network building
        net = tflearn.input_data([None, self.n_step, self.input_size, 1])
        net = tflearn.layers.normalization.batch_normalization(net)
        # layer1
        net = conv_2d(net, self.n_step * 2, int(self.input_size / 2), strides=2, activation='relu', name='Conv2D_1')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)
        # layer2
        net = conv_2d(net, self.n_step * 4, int(self.input_size / 4), strides=2, activation='relu', name='Conv2D_2')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)

        # layer2
        net = conv_2d(net, self.n_step * 8, int(self.input_size / 8), strides=2, activation='relu', name='Conv2D_3')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)

        # layer2
        net = conv_2d(net, self.n_step * 16, int(self.input_size / 16), strides=2, activation='relu', name='Conv2D_4')
        net = max_pool_2d(net, 3, strides=2)
        net = local_response_normalization(net)

        net = fully_connected(net, self.output_size, activation='softmax')
        net = tflearn.regression(net, optimizer='adam', learning_rate=0.001, loss='categorical_crossentropy')

        # Training
        _model = tflearn.DNN(net, tensorboard_verbose=3, checkpoint_path=self.checkpoint_path,
                             tensorboard_dir=self.tensorboard_dir)
        return _model

    def train(self, md_df, predict_test_random_state):
        xs, ys = self.get_x_y(md_df)
        trade_date_from, trade_date_to = md_df.index[0], md_df.index[-1]
        # xs_train, xs_validation, ys_train, ys_validation = self.separate_train_validation(xs, ys)
        if self.predict_test_random_state is None:
            random_state = predict_test_random_state
        else:
            random_state = self.predict_test_random_state

        xs_train, xs_validation, ys_train, ys_validation = train_test_split(
            xs, ys, test_size=0.2, random_state=random_state)
        self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = xs_train, xs_validation, ys_train, ys_validation
        # model.fit(trainX, trainY, validation_set=(testX, testY), show_metric=True, batch_size=32)
        sess = self.get_session(renew=True)
        train_acc, val_acc = 0, 0
        with sess.as_default():
            # with tf.Graph().as_default():
            # logger.debug('sess.graph:%s tf.get_default_graph():%s', sess.graph, tf.get_default_graph())
            tflearn.is_training(True)
            logger.debug('xs_train %s, ys_train %s, xs_validation %s, ys_validation %s [%s, %s]',
                         xs_train.shape, ys_train.shape, xs_validation.shape, ys_validation.shape,
                         trade_date_from, trade_date_to)
            for num in range(1, 6):
                logger.info('第 %d 轮训练开始 [%s, %s]', num, trade_date_from, trade_date_to)
                self.model.fit(xs_train, ys_train, validation_set=(xs_validation, ys_validation),
                               show_metric=False, batch_size=self.batch_size, n_epoch=self.n_epoch)

                result = self.model.evaluate(xs_validation, ys_validation, batch_size=self.batch_size)
                # logger.info("validation accuracy: %.2f%%" % (result[0] * 100))
                val_acc = result[0]
                if result[0] > 0.55:
                    break
                elif num < 5:
                    logger.warning('第 %d 轮训练，样本外训练精度不足，继续训练 [%s, %s]', num, trade_date_from, trade_date_to)

            tflearn.is_training(False)

            result = self.model.evaluate(xs_train, ys_train, batch_size=self.batch_size)
            # logger.info("train accuracy: %.2f%%" % (result[0] * 100))
            train_acc = result[0]

        # trade_date = md_df['trade_date'].iloc[-1]
        trade_date = md_df.index[-1]
        self.trade_date_last_train = trade_date
        return self.model, (train_acc, val_acc)

    def save_model(self):
        """
        将模型导出到文件
        :return:
        """
        # saver = tf.train.Saver(tf.trainable_variables(), max_to_keep=5)
        # save_path = saver.save(self.session, self.model_file_path)
        self.model.save(self.model_file_path)
        logger.info("模型训练截止日期： %s 保存到: %s", self.trade_date_last_train, self.model_file_path)
        return self.model_file_path

    def load_model_if_exist(self):
        """
        将模型导出到文件
        :return:
        """
        if self.enable_load_model_if_exist and self.model_file_exists():
            # 检查文件是否存在
            model = self.model  # 这句话是必须的，需要实现建立模型才可以加载
            # sess = self.session
            # saver = tf.train.Saver(tf.trainable_variables())
            # save_path = saver.restore(sess, self.model_file_path)
            model.load(self.model_file_path)
            logger.info("load from path: %s", self.model_file_path)
            return True

        return False

    def predict_test(self, md_df):
        logger.info('开始预测')
        # sess = self.session
        if self.xs_train is None:
            xs, ys = self.get_x_y(md_df)
            xs_train, xs_validation, ys_train, ys_validation = train_test_split(xs, ys, test_size=0.2, random_state=1)
            self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = xs_train, xs_validation, ys_train, ys_validation

        with self.session.as_default() as sess:
            # logger.info('sess.graph:%s', sess.graph)
            # logger.info('tf.get_default_graph():%s', tf.get_default_graph())
            # logger.info("批量预测")
            # result = self.model.evaluate(self.xs_validation, self.ys_validation, batch_size=self.batch_size)
            # logger.info("accuracy: %.2f%%" % (result[0] * 100))

            logger.info("批量预测2")
            real_ys = np.argmax(self.ys_validation, axis=1)
            # pred_ys = np.argmax(self.model.predict_label(self.xs_validation), axis=1) 与 evaluate 结果刚好相反
            # 因此不使用 predict_label 函数
            pred_ys = np.argmax(self.model.predict(self.xs_validation), axis=1)
            logger.info("accuracy: %.2f%%" % (sum(pred_ys == real_ys) / len(pred_ys) * 100))
            # logger.info("pred/real: \n%s\n%s", pred_ys, real_ys)

            logger.info("独立样本预测(predict_latest)")
            pred_ys = []
            for idx, y in enumerate(self.ys_validation):
                x = self.xs_validation[idx:idx + 1, :, :]
                pred_y = self.model.predict(x)
                pred_ys.extend(np.argmax(pred_y, axis=1))

            pred_ys = np.array(pred_ys)
            logger.info("accuracy: %.2f%%" % (sum(pred_ys == real_ys) / len(pred_ys) * 100))
            # logger.info("pred: \n%s\n%s", pred_ys, real_ys)

    def predict_latest(self, md_df):
        """
        计算最新一个 X，返回分类结果
        二分类，返回 0 / 1
        三分类，返回 0 / 1 / 2
        :param md_df:
        :return:
        """
        factors = self.get_factor_array(md_df, tail_n=self.n_step)
        x = self.get_batch_xs(factors)
        pred_y = np.argmax(self.model.predict(x), axis=1)[-1]
        # is_buy, is_sell = pred_mark == 1, pred_mark == 0
        # return is_buy, is_sell
        return pred_y

    def on_prepare_min1(self, md_df, context):
        if md_df is None:
            return
        df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        df.index = pd.DatetimeIndex(df.index)
        df = get_factor(df, close_key='close',
                        trade_date_series=self.trade_date_series,
                        delivery_date_series=self.delivery_date_series)

        self.load_train_test(df, enable_load_model=self.enable_load_model_if_exist)

    def load_train_test(self, md_df, enable_load_model, rebuild_model=False, enable_train_if_load_not_suss=True,
                        enable_train=True, enable_test=True):
        if rebuild_model:
            self.get_model(rebuild_model=True)

        # 加载模型
        if enable_load_model:
            is_load = self.load_model_if_exist()
        else:
            is_load = False

        if enable_train or not (enable_train_if_load_not_suss and not is_load):
            num = 0
            while True:
                num += 1
                # 训练模型
                _, (train_acc, val_acc) = self.train(md_df, num)
                if val_acc < 0.55:
                    logger.warning('第 %d 次训练，训练结果不及预期，重新采样训练', num)
                elif train_acc - val_acc > 0.1 or val_acc > 0.75:
                    logger.warning('第 %d 次训练，train_acc=%.4f, val_acc=%.4f 相差大于 10%%，重新采样训练', num, train_acc, val_acc)
                else:
                    break

            self.save_model()

            # trade_date = md_df['trade_date'].iloc[-1]
            trade_date = md_df.index[-1]
            self.trade_date_last_train = trade_date
        elif is_load:
            # trade_date = md_df['trade_date'].iloc[-1]
            trade_date = md_df.index[-1]
            self.trade_date_last_train = trade_date

        if enable_test:
            self.predict_test(md_df)

    def on_min1(self, md_df, context):
        # 数据整理
        df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        df.index = pd.DatetimeIndex(df.index)
        df = get_factor(df, close_key='close',
                        trade_date_series=self.trade_date_series,
                        delivery_date_series=self.delivery_date_series)
        # 获取最新jiaoyhiriq
        # trade_date = md_df['trade_date'].iloc[-1]
        trade_date = df.index[-1]
        days_after_last_train = (trade_date - self.trade_date_last_train).days
        if days_after_last_train >= self.retrain_period:
            # 重新训练
            logger.info('当前日期 %s 距离上一次训练 %s 已经过去 %d 天',
                        trade_date, self.trade_date_last_train, days_after_last_train)
            self.load_train_test(df, rebuild_model=True, enable_load_model=False, enable_train=True)

        # 预测
        pred_mark = self.predict_latest(df)
        is_buy, is_sell = pred_mark == 1, pred_mark == 0
        # logger.info('%s is_buy=%s, is_sell=%s', trade_date, str(is_buy), str(is_sell))
        close = md_df['close'].iloc[-1]
        instrument_id = context[ContextKey.instrument_id_list][0]
        if is_buy:  # is_buy
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Short:
                        self.close_short(instrument_id, close, pos_info.position)
                    elif direction == Direction.Long:
                        no_target_position = False
            if no_target_position:
                self.open_long(instrument_id, close, self.unit)
            else:
                logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_sell:  # is_sell
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Long:
                        self.close_long(instrument_id, close, pos_info.position)
                    elif direction == Direction.Short:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_short(instrument_id, close, self.unit)
            else:
                logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

    def model_file_exists(self):
        folder_path, file_name = os.path.split(self.model_file_path)
        if not os.path.exists(folder_path):
            return False
        for f_name in os.listdir(folder_path):
            if f_name.find(file_name) == 0:
                return True

        return False


def _test_use(is_plot):
    from ibats_common import module_root_path
    import os
    instrument_type = 'RB'
    # 参数设置
    run_mode = RunMode.Backtest_FixPercent
    calc_mode = CalcMode.Normal
    strategy_params = {'instrument_type': instrument_type, 'unit': 1}
    md_agent_params_list = [{
        'md_period': PeriodType.Min1,
        'instrument_id_list': [instrument_type],
        'datetime_key': 'trade_date',
        'init_md_date_from': '1995-1-1',  # 行情初始化加载历史数据，供策略分析预加载使用
        'init_md_date_to': '2013-1-1',
        # 'C:\GitHub\IBATS_Common\ibats_common\example\ru_price2.csv'
        'file_path': os.path.abspath(os.path.join(module_root_path, 'example', 'data', f'{instrument_type}.csv')),
        'symbol_key': 'instrument_type',
    }]
    if run_mode == RunMode.Realtime:
        trade_agent_params = {
        }
        strategy_handler_param = {
        }
    elif run_mode == RunMode.Backtest:
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            'init_cash': 10000,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2013-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    else:
        trade_agent_params = {
            'trade_mode': BacktestTradeMode.Order_2_Deal,
            "calc_mode": calc_mode,
        }
        strategy_handler_param = {
            'date_from': '2013-1-1',  # 策略回测历史数据，回测指定时间段的历史行情
            'date_to': '2018-10-18',
        }
    # 初始化策略处理器
    stghandler = strategy_handler_factory(
        stg_class=AIStg,
        strategy_params=strategy_params,
        md_agent_params_list=md_agent_params_list,
        exchange_name=ExchangeName.LocalFile,
        run_mode=run_mode,
        trade_agent_params=trade_agent_params,
        strategy_handler_param=strategy_handler_param,
    )
    stghandler.start()
    time.sleep(10)
    stghandler.keep_running = False
    stghandler.join()
    stg_run_id = stghandler.stg_run_id
    logging.info("执行结束 stg_run_id = %d", stg_run_id)

    if is_plot:
        from ibats_common.analysis.summary import summary_stg_2_docx
        from ibats_utils.mess import open_file_with_system_app
        file_path = summary_stg_2_docx(stg_run_id)
        open_file_with_system_app(file_path)

    return stg_run_id


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG, format=config.LOG_FORMAT)
    is_plot = True
    _test_use(is_plot)



# -*- coding: utf-8 -*-
"""
Created on 2017/11/13
@author: MG
"""
import logging
from enum import IntEnum, unique
from ibats_common.common import PeriodType
from ibats_common.config import config

logger = logging.getLogger()


@unique
class EventType(IntEnum):
    Tick_MD_Event = 0
    Min1_MD_Event = 1
    Min5_MD_Event = 5
    Min15_MD_Event = 15

    @staticmethod
    def try_2_period_type(event_type):
        """
        将 EventType类型转换为 PeriodType类型
        由于两个对象不一定一一对应，因此，无法匹配的类型返回None
        :param event_type:
        :return:
        """
        if event_type == EventType.Tick_MD_Event:
            period_type = PeriodType.Tick
        elif event_type == EventType.Min1_MD_Event:
            period_type = PeriodType.Min1
        elif event_type == EventType.Min5_MD_Event:
            period_type = PeriodType.Min5
        elif event_type == EventType.Min15_MD_Event:
            period_type = PeriodType.Min15
        else:
            # raise ValueError('event_type:%s 不是有效的类型' % event_type)
            logger.warning('event_type:%s 不是有效的类型', event_type)
            period_type = None
        return period_type


class EventAgent:
    """
    事件注册器，用于集中通过事件具备 处理相关类型的事件
    """

    def __init__(self):
        self._event_handler_dic = {}
        self._event_key_handler_dic = {}
        self.logger = logging.getLogger(self.__class__.__name__)

    def register_handler(self, event_type, handler, handler_name=None, key=None):
        """
        注册器:注册时间类型及对应的事件处理程序
        Key!=None，则该句柄只响应匹配 key 的事件
        Key=None，公共事件响应句柄，所有相应类型事件均响应
        :param event_type: 
        :param handler: 
        :param handler_name: 事件句柄名称，每个事件中，句柄名称唯一，重复将被覆盖
        :param key: 事件匹配 key，None代表公共事件响应句柄，所有该类型事件均响应
        :return: 
        """
        if event_type not in self._event_handler_dic:
            self._event_handler_dic[event_type] = {}
            self._event_key_handler_dic[event_type] = {}
        if handler_name is None:
            handler_name = handler.__name__
        self.logger.debug("注册事件处理句柄 %s -> <%s>%s",
                          event_type, handler_name, (" [%s]" % key) if key is not None else "")
        if key is None:
            if handler_name in self._event_handler_dic[event_type]:
                self.logger.warning('相应事件处理句柄 %s -> <%s> 已经存在，重新注册将覆盖原来执行句柄',
                                    event_type, handler_name)
            self._event_handler_dic[event_type][handler_name] = handler
        else:
            if key not in self._event_key_handler_dic[event_type]:
                self._event_key_handler_dic[event_type][key] = {}
            self._event_key_handler_dic[event_type][key][handler_name] = handler

    def send_event(self, event_type, data, key=None):
        """
        触发事件
        如果带Key，则触发 匹配 key 事件，以及全部 无 key 的事件响应句柄
        否则，只触发公共事件响应句柄
        :param event_type: 
        :param data: 
        :param key: 事件匹配 key，None代表只触发公共事件响应句柄
        :return: 
        """
        # 公共事件响应句柄
        if event_type in self._event_handler_dic:
            handler_dic = self._event_handler_dic[event_type]
            error_name_list = []
            for handler_name, handler in handler_dic.items():
                try:
                    handler(data)
                except:
                    self.logger.exception('%s run with error will be remove from register', handler_name)
                    error_name_list.append(handler_name)
            for handler_name in error_name_list:
                del handler_dic[handler_name]
                self.logger.warning('从注册器中移除 %s - %s', event_type, handler_name)
        # key 匹配事件响应句柄
        if key is not None and event_type in self._event_key_handler_dic:
            key_handler_dic = self._event_key_handler_dic[event_type]
            if key in key_handler_dic:
                handler_dic = key_handler_dic[key]
                error_name_list = []
                for handler_name, handler in handler_dic.items():
                    try:
                        handler(data)
                    except:
                        self.logger.exception('%s run with error will be remove from register', handler_name)
                        error_name_list.append(handler_name)
                for handler_name in error_name_list:
                    del handler_dic[handler_name]
                    self.logger.warning('从注册器中移除 %s - %s - %s', event_type, str(key), handler_name)


event_agent = EventAgent()

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format=config.LOG_FORMAT)
    event_agent.register_handler(EventType.Tick_MD_Event, lambda x: print('T:', x), 'print_t')
    event_agent.register_handler(EventType.Min1_MD_Event, lambda x: print('M:', x), 'print_m')
    event_agent.register_handler(EventType.Tick_MD_Event, lambda x: print('T 1:', x), 'print_t with key1', key='1')
    event_agent.send_event(EventType.Tick_MD_Event, 'sdf')
    event_agent.send_event(EventType.Tick_MD_Event, 'sdf', key='1')



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/3/8 12:23
@File    : orm_stg_run_status_detail_test.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_common.backend.orm import *
import unittest
from ibats_common.test.orm_trade_agent_detail_test import TradeAgentStatusDetailTest


class StgRunStatusDetailTest(unittest.TestCase):  # 继承unittest.TestCase
    def tearDown(self):
        pass

    def setUp(self):
        # 每个测试用例执行之前做操作
        pass

    @classmethod
    def tearDownClass(cls):
        # 必须使用 @ classmethod装饰器, 所有test运行完后运行一次
        pass

    @classmethod
    def setUpClass(cls):
        # from ibats_common.config import update_db_config, ConfigBase
        # update_db_config({
        #     ConfigBase.DB_SCHEMA_IBATS: 'mysql://mg:Dcba1234@localhost/' + ConfigBase.DB_SCHEMA_IBATS,
        # })
        # 必须使用@classmethod 装饰器,所有test运行前运行一次
        init()
        global engine_ibats
        engine_ibats = engines.engine_ibats

    def test_create_by_trade_agent_status_detail_list(self):
        info, status = TradeAgentStatusDetailTest.add_trade_agent_status_detail()
        detail = StgRunStatusDetail.create_by_trade_agent_status_detail_list(info.stg_run_id, [status])
        self.assertEqual(detail.stg_run_id, info.stg_run_id)
        self.assertEqual(detail.cash_available, status.cash_available)
        self.assertEqual(detail.position_value, status.position_value)
        self.assertEqual(detail.curr_margin, status.curr_margin)
        self.assertEqual(detail.close_profit, status.close_profit)
        self.assertEqual(detail.position_profit, status.position_profit)
        self.assertEqual(detail.floating_pl_cum, status.floating_pl_cum)
        self.assertEqual(detail.commission_tot, status.commission_tot)
        self.assertEqual(detail.cash_init, status.cash_init)
        self.assertEqual(detail.cash_and_margin, status.cash_and_margin)
        self.assertEqual(detail.cashflow, status.cashflow)
        self.assertEqual(detail.cashflow_cum, status.cashflow_cum)
        self.assertEqual(detail.rr, status.rr)


if __name__ == '__main__':
    unittest.main()  # 运行所有的测试用例



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/20 19:53
@File    : md_agent.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import time

from ibats_utils.mess import str_2_date

from ibats_common.md import MdAgentBase, md_agent
from ibats_common.common import PeriodType, RunMode, ExchangeName
import pandas as pd


class MdAgentPub(MdAgentBase):

    def __init__(self, instrument_id_list, md_period: PeriodType, exchange_name, file_path, agent_name=None,
                 init_load_md_count=None, init_md_date_from=None, init_md_date_to=None, ffill_on_load_history=True,
                 **kwargs):
        MdAgentBase.__init__(
            self, instrument_id_list, md_period, exchange_name, agent_name=agent_name,
            init_load_md_count=init_load_md_count, init_md_date_from=init_md_date_from,
            init_md_date_to=init_md_date_to, **kwargs)
        self.file_path = file_path
        if 'datetime_key' in kwargs:
            self.datetime_key = kwargs['datetime_key']
            self.date_key = None
            self.time_key = None
        else:
            self.datetime_key = None
            self.date_key = kwargs['date_key']
            self.time_key = kwargs['time_key']
        if 'microseconds_key' in kwargs:
            self.microseconds_key = kwargs['microseconds_key']
        else:
            self.microseconds_key = None

        self.factors = kwargs['factors'] if 'factors' in kwargs else None
        self.ffill_on_load_history = ffill_on_load_history

    def load_history(self, date_from=None, date_to=None, load_md_count=None) -> (pd.DataFrame, dict):
        """
        从文件中加载历史数据
        :param date_from: None代表沿用类的 init_md_date_from 属性
        :param date_to: None代表沿用类的 init_md_date_from 属性
        :param load_md_count: 0 代表不限制，None代表沿用类的 init_load_md_count 属性，其他数字代表相应的最大加载条数
        :return: md_df 或者
         ret_data {
            'md_df': md_df, 'datetime_key': 'ts_start',
            'date_key': **, 'time_key': **, 'microseconds_key': **
            }
        """
        # 如果 init_md_date_from 以及 init_md_date_to 为空，则不加载历史数据
        if self.init_md_date_from is None and self.init_md_date_to is None:
            ret_data = {'md_df': None, 'datetime_key': self.datetime_key, 'date_key': self.date_key,
                        'time_key': self.time_key, 'microseconds_key': self.microseconds_key}
            return ret_data

        # 加载历史数据
        self.logger.debug("加载历史数据 %s", self.file_path)
        date_col_name_set = {self.datetime_key, self.date_key, self.timestamp_key}
        if None in date_col_name_set:
            date_col_name_set.remove(None)
        md_df = pd.read_csv(self.file_path, parse_dates=list(date_col_name_set))
        if self.ffill_on_load_history:
            md_df = md_df.ffill()
        if self.timestamp_key is not None:
            timestamp_key = self.timestamp_key
        elif self.datetime_key is not None:
            timestamp_key = self.datetime_key
        elif self.date_key is not None:
            timestamp_key = self.date_key
        else:
            timestamp_key = None
            self.logger.warning('没有设置 timestamp_key、datetime_key、date_key 中的任何一个，无法进行日期过滤')

        # 对日期区间进行过滤
        filter_mark = None
        if timestamp_key is not None:
            if date_from is None:
                date_from = pd.Timestamp(str_2_date(self.init_md_date_from))
            else:
                date_from = pd.Timestamp(str_2_date(date_from))

            if date_from is not None:
                filter_mark = md_df[timestamp_key] >= date_from

            if date_to is None:
                date_to = pd.Timestamp(str_2_date(self.init_md_date_to))
            else:
                date_to = pd.Timestamp(str_2_date(date_to))

            if date_to is not None:
                if filter_mark is None:
                    filter_mark = md_df[timestamp_key] <= date_to
                else:
                    filter_mark &= md_df[timestamp_key] <= date_to

        if filter_mark is not None:
            ret_df = md_df[filter_mark]
        else:
            ret_df = md_df

        # 返回数据
        ret_data = {'md_df': ret_df, 'datetime_key': self.datetime_key, 'date_key': self.date_key,
                    'time_key': self.time_key, 'microseconds_key': self.microseconds_key,
                    'symbol_key': self.symbol_key, 'close_key': 'close'}
        return ret_data


@md_agent(RunMode.Backtest, ExchangeName.LocalFile, is_default=False)
class MdAgentBacktest(MdAgentPub):

    def __init__(self, **kwargs):
        MdAgentPub.__init__(self, **kwargs)
        self.timeout = 1
        self.timestamp_key = kwargs['timestamp_key'] if 'timestamp_key' in kwargs else self.datetime_key
        self.symbol_key = kwargs['symbol_key']
        self.close_key = kwargs['close_key'] if 'close_key' in kwargs else 'close'

    def connect(self):
        """链接redis、初始化历史数据"""
        pass

    def release(self):
        """释放channel资源"""
        pass

    def run(self):
        """启动多线程获取MD"""
        if not self.keep_running:
            self.keep_running = True
            while self.keep_running:
                time.sleep(self.timeout)
            else:
                self.logger.info('%s job finished', self.name)


@md_agent(RunMode.Backtest_FixPercent, ExchangeName.LocalFile, is_default=False)
class MdAgentBacktestFixPercent(MdAgentPub):

    def __init__(self, **kwargs):
        MdAgentPub.__init__(self, **kwargs)
        self.timeout = 1
        self.timestamp_key = kwargs['timestamp_key'] if 'timestamp_key' in kwargs else self.datetime_key
        self.symbol_key = kwargs['symbol_key']
        self.close_key = kwargs['close_key'] if 'close_key' in kwargs else 'close'

    def connect(self):
        """链接redis、初始化历史数据"""
        pass

    def release(self):
        """释放channel资源"""
        pass

    def run(self):
        """启动多线程获取MD"""
        if not self.keep_running:
            self.keep_running = True
            while self.keep_running:
                time.sleep(self.timeout)
            else:
                self.logger.info('%s job finished', self.name)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/3/8 12:23
@File    : orm_stg_run_status_detail_test.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_common.backend.orm import *
from datetime import date, datetime
import unittest

from ibats_common.common import ExchangeName


class InitTest(unittest.TestCase):  # 继承unittest.TestCase
    def tearDown(self):
        pass

    def setUp(self):
        # 每个测试用例执行之前做操作
        pass

    @classmethod
    def tearDownClass(cls):
        # 必须使用 @ classmethod装饰器, 所有test运行完后运行一次
        pass

    @classmethod
    def setUpClass(cls):
        # from ibats_common.config import update_db_config, ConfigBase
        # update_db_config({
        #     ConfigBase.DB_SCHEMA_IBATS: 'mysql://mg:Dcba1234@localhost/' + ConfigBase.DB_SCHEMA_IBATS,
        # })
        # 必须使用@classmethod 装饰器,所有test运行前运行一次
        init()
        global engine_ibats
        engine_ibats = engines.engine_ibats

    @staticmethod
    def add_stg_run_info():
        info = StgRunInfo()
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(info)
            session.commit()
        return info

    def test_add_stg_run_info(self):
        info = InitTest.add_stg_run_info()

        self.assertIsNotNone(info.stg_run_id)
        self.assertGreater(info.stg_run_id, -1)

        with with_db_session(engine_ibats) as session:
            info2 = session.query(StgRunInfo).filter(StgRunInfo.stg_run_id == info.stg_run_id).first()

        self.assertIsInstance(info2, StgRunInfo)
        self.assertEqual(info.stg_run_id, info2.stg_run_id)

    @staticmethod
    def add_order(stg_run_id=None):
        if stg_run_id is None:
            info = InitTest.add_stg_run_info()
            stg_run_id = info.stg_run_id

        order = OrderDetail(stg_run_id, trade_agent_key=ExchangeName.DataIntegration,
                            order_dt=datetime.now(), order_date=date.today(), order_time=datetime.now().time(),
                            order_millisec=99, direction=int(Direction.Long), action=int(Action.Open), symbol='RB1801',
                            order_price=1000.0, order_vol=20, calc_mode=CalcMode.Margin,
                            )
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(order)
            session.commit()
        return order

    def test_add_order_detail(self):
        order = InitTest.add_order()
        self.assertIsNotNone(order)
        self.assertGreater(order.order_idx, -1)

        with with_db_session(engine_ibats) as session:
            order2 = session.query(OrderDetail).filter(
                OrderDetail.order_idx == order.order_idx,
                OrderDetail.stg_run_id == order.stg_run_id
            ).first()

        self.assertIsInstance(order2, OrderDetail)
        self.assertEqual(order2.order_idx, order.order_idx)

    def test_add_trade_detail(self):
        trade = InitTest.add_trade()

        self.assertIsNotNone(trade)
        self.assertGreater(trade.trade_idx, -1)

        with with_db_session(engine_ibats) as session:
            trade2 = session.query(TradeDetail).filter(
                TradeDetail.trade_idx == trade.trade_idx,
                TradeDetail.stg_run_id == trade.stg_run_id
            ).first()

        self.assertIsInstance(trade2, TradeDetail)
        self.assertEqual(trade.trade_idx, trade2.trade_idx)

    @staticmethod
    def add_trade():
        order = InitTest.add_order()
        trade = TradeDetail(order.stg_run_id, trade_agent_key=ExchangeName.DataIntegration, order_idx=order.order_idx,
                            order_price=1000.0, order_vol=20,
                            trade_dt=datetime.now(), trade_date=date.today(), trade_time=datetime.now().time(),
                            trade_millisec=99, direction=int(Direction.Long), action=int(Action.Open), symbol='RB1801',
                            trade_price=order.order_price + 1, trade_vol=order.order_vol,
                            margin=order.order_price * order.order_vol,
                            commission=order.order_price * order.order_vol + 0.00005, multiple=10, margin_ratio=1
                            )
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(trade)
            session.commit()
        return trade

    def test_add_pos_status(self):
        pos_status = InitTest.add_pos_status()

        self.assertIsNotNone(pos_status)
        self.assertGreater(pos_status.pos_status_detail_idx, -1)

        with with_db_session(engine_ibats) as session:
            pos_status2 = session.query(PosStatusDetail).filter(
                PosStatusDetail.pos_status_detail_idx == pos_status.pos_status_detail_idx,
                PosStatusDetail.stg_run_id == pos_status.stg_run_id
            ).first()

        self.assertIsInstance(pos_status2, PosStatusDetail)
        self.assertEqual(pos_status.pos_status_detail_idx, pos_status2.pos_status_detail_idx)

    @staticmethod
    def add_pos_status():
        trade = InitTest.add_trade()
        commission_rate = 0.0005
        multiple, margin_ratio = 10, 0.12
        pos_status = PosStatusDetail(
            trade.stg_run_id, trade_agent_key=trade.trade_agent_key, trade_idx=trade.trade_idx,
            trade_dt=trade.trade_dt, trade_date=trade.trade_date, trade_time=trade.trade_time,
            trade_millisec=trade.trade_millisec, direction=trade.direction,
            symbol=trade.symbol,
            position=trade.trade_vol, position_chg=-trade.trade_vol,
            avg_price=trade.trade_price,
            cur_price=trade.trade_price, floating_pl=-1, floating_pl_rate=-commission_rate,
            floating_pl_chg=trade.trade_price * trade.trade_vol * trade.multiple * commission_rate,
            floating_pl_cum=-13, cashflow=123, cashflow_cum=1234, rr=-commission_rate,
            margin=trade.trade_price * trade.trade_vol,
            margin_chg=-trade.trade_price * trade.trade_vol,
            position_date_type=PositionDateType.Today.value,
            commission=trade.trade_vol * trade.trade_price * commission_rate * trade.multiple,
            commission_tot=trade.trade_vol * trade.trade_price * commission_rate * trade.multiple,
            multiple=multiple,
            margin_ratio=margin_ratio,
            calc_mode=CalcMode.Margin.value
        )
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(pos_status)
            session.commit()
        return pos_status

    def test_add_trade_agent_status(self):
        status = InitTest.add_trade_agent_status()

        self.assertIsNotNone(status)
        self.assertGreater(status.trade_agent_status_detail_idx, -1)

        with with_db_session(engine_ibats) as session:
            status2 = session.query(TradeAgentStatusDetail).filter(
                TradeAgentStatusDetail.trade_agent_status_detail_idx == status.trade_agent_status_detail_idx,
                TradeAgentStatusDetail.stg_run_id == status.stg_run_id
            ).first()

        self.assertIsInstance(status2, TradeAgentStatusDetail)
        self.assertEqual(status.trade_agent_status_detail_idx, status2.trade_agent_status_detail_idx)

    @staticmethod
    def add_trade_agent_status():
        pos_status = InitTest.add_pos_status()
        cash_init = 100000
        cash_available = cash_init - pos_status.margin - pos_status.commission
        status = TradeAgentStatusDetail(
            pos_status.stg_run_id, trade_agent_key=pos_status.trade_agent_key,
            trade_dt=pos_status.trade_dt, trade_date=pos_status.trade_date,
            trade_time=pos_status.trade_time,
            trade_millisec=pos_status.trade_millisec,
            cash_available=cash_available,
            curr_margin=pos_status.margin,
            close_profit=0, position_profit=pos_status.floating_pl,
            floating_pl_cum=pos_status.floating_pl_cum,
            commission_tot=pos_status.commission_tot,
            cash_init=cash_init
            )
        status.cash_and_margin_value = cash_available + pos_status.margin
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(status)
            session.commit()
        return status

    def test_add_stg_run_detail(self):
        detail = InitTest.add_stg_run_detail()

        self.assertIsNotNone(detail)
        self.assertGreater(detail.stg_run_status_detail_idx, -1)

        with with_db_session(engine_ibats) as session:
            detail2 = session.query(StgRunStatusDetail).filter(
                StgRunStatusDetail.stg_run_status_detail_idx == detail.stg_run_status_detail_idx,
                StgRunStatusDetail.stg_run_id == detail.stg_run_id
            ).first()

        self.assertIsInstance(detail2, StgRunStatusDetail)
        self.assertEqual(detail.stg_run_status_detail_idx, detail2.stg_run_status_detail_idx)

    @staticmethod
    def add_stg_run_detail():
        status = InitTest.add_trade_agent_status()
        detail = StgRunStatusDetail(
            status.stg_run_id, trade_dt=status.trade_dt, trade_date=status.trade_date,
            trade_time=status.trade_time,
            trade_millisec=status.trade_millisec, cash_available=0,
            curr_margin=status.curr_margin,
            close_profit=status.close_profit, position_profit=status.position_profit,
            floating_pl_cum=status.floating_pl_cum,
            commission_tot=status.commission_tot, cash_init=status.cash_init,
            cashflow_daily=status.cashflow_daily,
            cashflow_cum=status.cashflow_cum,
            rr=status.rr,
        )
        with with_db_session(engine_ibats, expire_on_commit=False) as session:
            session.add(detail)
            session.commit()
        return detail


if __name__ == '__main__':
    unittest.main()  # 运行所有的测试用例



# -*- coding: utf-8 -*-
"""
Created on 2017/6/9
@author: MG
"""
import logging
from logging.config import dictConfig
from datetime import datetime

logger = logging.getLogger(__name__)


class ConfigBase:

    DB_SCHEMA_IBATS = 'ibats'
    DB_URL_DIC = {
        DB_SCHEMA_IBATS: 'mysql://mg:Abcd1234@localhost/' + DB_SCHEMA_IBATS,
    }

    BACKTEST_UPDATE_OR_INSERT_PER_ACTION = False
    ORM_UPDATE_OR_INSERT_PER_ACTION = True
    UPDATE_STG_RUN_STATUS_DETAIL_PERIOD = 1  # 1 每一个最小行情周期，2 每天

    # evn configuration
    LOG_FORMAT = '%(asctime)s %(levelname)s %(name)s %(filename)s.%(funcName)s:%(lineno)d|%(message)s'

    # log settings
    logging_config = dict(
        version=1,
        formatters={
            'simple': {
                'format': LOG_FORMAT}
        },
        handlers={
            'file_handler':
                {
                    'class': 'logging.handlers.RotatingFileHandler',
                    'filename': 'logger.log',
                    'maxBytes': 1024 * 1024 * 10,
                    'backupCount': 5,
                    'level': 'DEBUG',
                    'formatter': 'simple',
                    'encoding': 'utf8'
                },
            'console_handler':
                {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG',
                    'formatter': 'simple'
                }
        },

        root={
            'handlers': ['console_handler', 'file_handler'],
            'level': logging.DEBUG,
        }
    )
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARN)
    logging.getLogger('urllib3.connectionpool').setLevel(logging.INFO)
    logging.getLogger('matplotlib').setLevel(logging.INFO)
    dictConfig(logging_config)
    update_db_datetime = datetime.now()


# 开发配置（SIMNOW MD + Trade）
config = ConfigBase()


def update_config(config_new: ConfigBase):
    """更新配置信息"""
    global config
    config = config_new
    config.update_db_datetime = datetime.now()
    logger.info('更新默认配置信息 %s < %s', ConfigBase, config_new.__class__)


def update_db_config(db_url_dic: dict):
    """更新数据配置链接"""
    config.DB_URL_DIC.update(db_url_dic)
    logger.debug('更新数据库配置信息  %s keys: %s', ConfigBase, list(db_url_dic.keys()))
    config.update_db_datetime = datetime.now()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/11/7 10:47
@File    : strategy_handler.py
@contact : mmmaaaggg@163.com
@desc    : 策略处理句柄，用于处理策略进行回测或实盘交易
"""
import json
import logging
import time
import warnings
from abc import ABC
from collections import defaultdict
from datetime import date, datetime
from functools import lru_cache
from queue import Empty
from threading import Thread

from ibats_utils.db import with_db_session
from ibats_utils.mess import try_2_date, load_class, get_module_path
from sqlalchemy.exc import SQLAlchemyError

from ibats_common.backend import engines
from ibats_common.backend.orm import StgRunInfo, StgRunStatusDetail
from ibats_common.common import ExchangeName, RunMode, ContextKey, CalcMode
from ibats_common.config import config
from ibats_common.md import md_agent_factory
from ibats_common.strategy import StgBase
from ibats_common.trade import trader_agent_factory

engine_ibats = engines.engine_ibats
logger = logging.getLogger(__name__)


class StgHandlerBase(Thread, ABC):

    def __init__(self, stg_run_id, stg_base: StgBase, run_mode, md_key_period_agent_dic):
        super().__init__(daemon=True)
        self.stg_run_id = stg_run_id
        self.run_mode = run_mode
        # 初始化策略实体，传入参数
        self.stg_base = stg_base
        # 设置工作状态
        self.is_working = False
        self.is_done = False
        # 日志
        self.logger = logging.getLogger(str(self.__class__))
        # 对不同周期设置相应的md_agent
        self.md_key_period_agent_dic = md_key_period_agent_dic
        self.stg_run_status_detail_list = []

    def stg_run_ending(self):
        """
        处理策略结束相关事项
        释放策略资源
        更新策略执行信息
        :return:
        """
        self.stg_base.release()
        # 更新数据库 td_to 字段
        with with_db_session(engine_ibats) as session:
            session.query(StgRunInfo).filter(StgRunInfo.stg_run_id == self.stg_run_id).update(
                {StgRunInfo.dt_to: datetime.now()})
            try:
                session.bulk_save_objects(self.stg_run_status_detail_list)
                # sql_str = StgRunInfo.update().where(
                # StgRunInfo.c.stg_run_id == self.stg_run_id).values(dt_to=datetime.now())
                # session.execute(sql_str)
                session.commit()
                self.logger.debug("%d 条 stg_run_status_detail 被保存", len(self.stg_run_status_detail_list))
            except SQLAlchemyError:
                logger.exception("%d 条 stg_run_status_detail 被保存时发生异常", len(self.stg_run_status_detail_list))
                session.rollback()

        self.is_working = False
        self.is_done = True

    def __repr__(self):
        return '<{0.__class__.__name__}:{0.stg_run_id} {0.run_mode}>'.format(self)


class StgHandlerRealtime(StgHandlerBase):

    def __init__(self, stg_run_id, stg_base: StgBase, md_key_period_agent_dic, **kwargs):
        super().__init__(stg_run_id=stg_run_id, stg_base=stg_base, run_mode=RunMode.Realtime,
                         md_key_period_agent_dic=md_key_period_agent_dic)
        # 对不同周期设置相应的md_agent
        self.md_period_agent_dic = md_key_period_agent_dic
        # 设置线程池
        self.running_thread = {}
        # 设置推送超时时间
        self.timeout_pull = 60
        # 设置独立的时间线程
        self.enable_timer_thread = kwargs.setdefault('enable_timer_thread', False)
        self.seconds_of_timer_interval = kwargs.setdefault('seconds_of_timer_interval', 9999)

    def run(self):

        # TODO: 以后再加锁，防止多线程，目前只是为了防止误操作导致的重复执行
        if self.is_working:
            return
        else:
            self.is_working = True

        try:
            # 策略初始化
            self.stg_base.init()
            # 对各个周期分别设置对应 handler
            for period, md_agent in self.md_period_agent_dic.items():
                # 获取对应事件响应函数
                on_period_md_handler = self.stg_base.on_period_md_handler
                # 异步运行：每一个周期及其对应的 handler 作为一个线程独立运行
                thread_name = 'run_md_agent %s' % md_agent.name
                run_md_agent_thread = Thread(target=self.run_md_agent, name=thread_name,
                                             args=(md_agent, on_period_md_handler), daemon=True)
                self.running_thread[period] = run_md_agent_thread
                self.logger.info("加载 %s 线程", thread_name)
                run_md_agent_thread.start()

            if self.enable_timer_thread:
                thread_name = 'run_timer'
                timer_thread = Thread(target=self.run_timer, name=thread_name, daemon=True)
                self.logger.info("加载 %s 线程", thread_name)
                timer_thread.start()

            # 各个线程分别join等待结束信号
            for period, run_md_agent_thread in self.running_thread.items():
                run_md_agent_thread.join()
                self.logger.info('%s period %s finished', run_md_agent_thread.name, period)
        finally:
            self.is_working = False
            self.stg_run_ending()

    def run_timer(self):
        """
        负责定时运行策略对象的 on_timer 方法
        :return:
        """
        while self.is_working:
            try:
                self.stg_base.on_timer()
            except:
                self.logger.exception('on_timer 函数运行异常')
            finally:
                time.sleep(self.seconds_of_timer_interval)

    def run_md_agent(self, md_agent, handler):
        """
        md_agent pull 方法的事件驱动处理函数
        :param md_agent:
        :param handler:  self.stgbase对象的响应 md_agent 的梳理函数：根据不同的 md_period 可能是
         on_tick、 on_min、 on_day、 on_week、 on_month 等其中一个
        :return:
        """
        period = md_agent.md_period
        self.logger.info('启动 %s 行情监听线程', period)
        md_agent.connect()
        md_agent.subscribe()  # 参数为空相当于 md_agent.subscribe(md_agent.instrument_id_list)
        md_agent.start()
        md_dic = None
        while self.is_working:
            try:
                if not self.is_working:
                    break
                # 加载数据，是设置超时时间，防止长时间阻塞
                md_dic = md_agent.pull(self.timeout_pull)
                handler(period, md_dic)
            except Empty:
                # 工作状态检查
                pass
            except Exception:
                self.logger.exception('%s 事件处理句柄执行异常，对应行情数据md_dic:\n%s',
                                      period, md_dic)
                # time.sleep(1)
        md_agent.release()
        self.logger.info('period:%s finished', period)


class StgHandlerBacktest(StgHandlerBase):

    def __init__(self, stg_run_id, stg_base: StgBase, run_mode, md_key_period_agent_dic, date_from, date_to,
                 md_td_agent_key_list_map=None, **kwargs):
        super().__init__(stg_run_id=stg_run_id, stg_base=stg_base, run_mode=run_mode,
                         md_key_period_agent_dic=md_key_period_agent_dic)
        # 设置回测时间区间
        self.date_from = try_2_date(date_from)
        self.date_to = try_2_date(date_to)
        self.params = kwargs
        if not isinstance(self.date_from, date):
            raise ValueError("date_from: %s", date_from)
        if not isinstance(self.date_to, date):
            raise ValueError("date_from: %s", date_to)
        # 用于维护 md_agent 与 td_agent 之间对应关系（多对多关系）
        # 用于在回测成交时指定行情发送到哪一个 trade_agent，默认情况下，该map将交易所相同的 md td agent进行匹配，也可以根据需要在参数设是手动进行配置
        self.logger.debug('设置 md_td_agent_key_list_map')
        if md_td_agent_key_list_map is not None:
            self.md_td_agent_key_list_map = md_td_agent_key_list_map
        else:
            self.md_td_agent_key_list_map = defaultdict(list)
            # 遍历所有 md_agent, 查找 exchange_name 相同的进行 mapping
            for md_agent_key, period_agent_dic in self.md_key_period_agent_dic.items():
                for period, md_agent in period_agent_dic.items():
                    for trade_agent_key, trade_agent in self.stg_base.trade_agent_dic.items():
                        if md_agent.exchange_name == trade_agent.exchange_name:
                            self.md_td_agent_key_list_map[md_agent_key].append(trade_agent_key)

                    break

        # 设置相应 trade_agent 的 关键 key 信息
        self.logger.debug('设置 关键 key 信息')
        for md_agent_key, trade_agent_key_list in self.md_td_agent_key_list_map.items():
            for period, md_agent in self.md_key_period_agent_dic[md_agent_key].items():
                for trade_agent_key in trade_agent_key_list:
                    md_agent.check_key()
                    self.stg_base.trade_agent_dic[trade_agent_key].set_timestamp_key(md_agent.timestamp_key)
                    self.stg_base.trade_agent_dic[trade_agent_key].set_symbol_key(md_agent.symbol_key)
                    self.stg_base.trade_agent_dic[trade_agent_key].set_close_key(md_agent.close_key)

                break

    def get_periods_history_iterator(self) -> dict:
        """
        获取各个周期的 历史数据的迭代器
        :return:
        """
        md_agent_key_cor_func_dic = {}
        # 每一个 md_agent 抓取第一条记录，放入 md_list_sorted_by_datetime_tag 并按 datetime_tag 排序
        for md_agent_key, period_agent_dic in self.md_key_period_agent_dic.items():
            for period, md_agent in period_agent_dic.items():
                cor_func = md_agent.cor_load_history_record(self.date_from, self.date_to, load_md_count=0)
                meta_dic = {
                    'symbol_key': md_agent.symbol_key,
                    'close_key': md_agent.close_key,
                    'timestamp_key': md_agent.timestamp_key
                }
                md_agent_key_cor_func_dic[(md_agent_key, period)] = (cor_func, meta_dic)
        return md_agent_key_cor_func_dic

    def load_history_record(self):
        """
        迭代器方法，用于产生行情数据
        :return:
        """
        md_list_sorted_by_datetime_tag = []
        # 每一个 md_agent 抓取第一条记录，放入 md_list_sorted_by_datetime_tag 并按 datetime_tag 排序
        for md_agent_key, period_agent_dic in self.md_key_period_agent_dic.items():
            for period, md_agent in period_agent_dic.items():
                cor_func = md_agent.cor_load_history_record(self.date_from, self.date_to, load_md_count=0)
                try:
                    num, datetime_tag, md_s = cor_func.send(None)
                except StopIteration as exp:
                    self.logger.info('%s %s 没有数据', md_agent_key, period)
                    continue

                md_list_sorted_by_datetime_tag.append((datetime_tag, md_agent_key, period, num, md_s, cor_func))

        # 开始循环，每一次循环：
        # 1）读取发送 datetime_tag 最小的一条记录
        # 2）从相应的 md_agent 再读取一条记录，补充到 md_list_sorted_by_datetime_tag
        # 3）按 datetime_tag 排序
        # 如此循环，知道全部数据被发送完
        md_list_sorted_by_datetime_tag.sort(key=lambda x: x[0])
        list_count = len(md_list_sorted_by_datetime_tag)
        data_count = 0
        while list_count > 0:
            datetime_tag, md_agent_key, period, num, md_s, cor_func = md_list_sorted_by_datetime_tag[0]
            yield datetime_tag, md_agent_key, period, num, md_s
            try:
                num, datetime_tag, md_s = cor_func.send(None)
                md_list_sorted_by_datetime_tag[0] = (datetime_tag, md_agent_key, period, num, md_s, cor_func)
            except StopIteration as exp:
                data_count += exp.value
                self.logger.info('%s %s 推送 %d 条数据完成', md_agent_key, period, exp.value)
                md_list_sorted_by_datetime_tag.pop(0)
                list_count = len(md_list_sorted_by_datetime_tag)

            # 重新按 datetime_tag 排序
            md_list_sorted_by_datetime_tag.sort(key=lambda x: x[0])

        self.logger.info('全部数据推送完成， 累计推送 %d 条数据', data_count)

    def _update_stg_run_status_detail(self, trade_agent_status_detail_list):
        if trade_agent_status_detail_list is not None and len(trade_agent_status_detail_list) > 0:
            if len(self.stg_run_status_detail_list) == 0:
                detail_last = StgRunStatusDetail.create_t_1(
                    self.stg_run_id, trade_agent_status_detail_list)
                self.stg_run_status_detail_list.append(detail_last)
            else:
                detail_last = self.stg_run_status_detail_list[-1]

            detail = detail_last.update_by_trade_agent_status_detail_list(
                trade_agent_status_detail_list)
            self.stg_run_status_detail_list.append(detail)

    def run(self):
        """
        执行回测
        :return:
        """
        # TODO: 以后再加锁，防止多线程，目前只是为了防止误操作导致的重复执行
        if self.is_working:
            self.logger.warning('当前任务正在执行中..，避免重复执行')
            return
        else:
            self.is_working = True
        self.logger.info('执行回测任务【%s - %s】开始', self.date_from, self.date_to)
        try:
            # 策略初始化
            self.stg_base.init()
            # 对每个 md_agent 激活 load_history_record
            # 每一次取记录包括：时间轴标签 + 行情数据
            # 比较各个 md_agent 取时间最小的作为本次数据推送，如此循环，直至所有 mg_agent 记录均提取完成
            # 按照时间顺序将各个周期数据依次推入对应 handler
            data_count = 0
            datetime_tag_last = None
            for data_count, (datetime_tag, md_agent_key, period, num, md_s) in enumerate(
                    self.load_history_record(), start=1):
                md = md_s.to_dict()
                # self.logger.debug("md: %s", md)
                # 在回测阶段，需要对 trade_agent 设置最新的md数据，一遍交易接口确认相应的k线日期
                for trade_agent_key in self.md_td_agent_key_list_map[md_agent_key]:
                    if trade_agent_key == ExchangeName.Default:
                        continue
                    self.stg_base.trade_agent_dic[trade_agent_key].set_curr_md(period, md)

                # 执行策略相应的事件响应函数
                self.stg_base.on_period_md_handler(period, md, md_agent_key)
                # 根据最新的 md 及 持仓信息 更新 账户信息
                for trade_agent_key in self.md_td_agent_key_list_map[md_agent_key]:
                    if trade_agent_key == ExchangeName.Default:
                        continue
                    self.stg_base.trade_agent_dic[trade_agent_key].update_trade_agent_status_detail()

                # 以下两种情况开始计算 StgRunStatusDetail：
                # 1）config.UPDATE_STG_RUN_STATUS_DETAIL_PERIOD == 1 时，每一次状态更新均重新计算
                # 2）每一个自然日开始
                # 首次运行的时候，回触发一次初始化，先生成一条 T-1日的状态记录
                if config.UPDATE_STG_RUN_STATUS_DETAIL_PERIOD == 1:
                    trade_agent_status_detail_list = [
                        trade_agent.trade_agent_status_detail_latest
                        for key, trade_agent in self.stg_base.trade_agent_dic.items()
                        if key != ExchangeName.Default and trade_agent.trade_agent_status_detail_latest is not None]

                elif (config.UPDATE_STG_RUN_STATUS_DETAIL_PERIOD == 2
                      and datetime_tag_last is not None
                      and datetime_tag_last.date() != datetime_tag.date()):
                    # TODO: 这里应该是找前一交易日的最后一个 trade_agent_status_detail
                    trade_agent_status_detail_list = [
                        trade_agent.trade_agent_status_detail_latest
                        for key, trade_agent in self.stg_base.trade_agent_dic.items()
                        if key != ExchangeName.Default and trade_agent.trade_agent_status_detail_latest is not None]
                else:
                    trade_agent_status_detail_list = None

                self._update_stg_run_status_detail(trade_agent_status_detail_list)

                datetime_tag_last = datetime_tag

            # 循环结束
            self.logger.info('执行回测任务【%s - %s】完成，处理数据 %d 条', self.date_from, self.date_to, data_count)
        finally:
            if config.UPDATE_STG_RUN_STATUS_DETAIL_PERIOD == 2:
                trade_agent_status_detail_list = [
                    trade_agent.trade_agent_status_detail_latest
                    for trade_agent in self.stg_base.trade_agent_dic.values()
                    if trade_agent is not None]
                self._update_stg_run_status_detail(trade_agent_status_detail_list)

            self.is_working = False
            self.stg_run_ending()


def strategy_handler_factory(
        stg_class: type(StgBase), strategy_params, md_agent_params_list, run_mode: RunMode, exchange_name: ExchangeName,
        trade_agent_params: dict, strategy_handler_param: dict) -> StgHandlerBase:
    """
    单一交易所策略处理具备
    建立策略对象
    建立数据库相应记录信息
    根据运行模式（实时、回测）：选择相应的md_agent以及trade_agent
    :param stg_class: 策略类型 StgBase 的子类
    :param strategy_params: 策略参数
    :param md_agent_params_list: 行情代理（md_agent）参数，支持同时订阅多周期、多品种，
    例如：同时订阅 [ethusdt, eosusdt] 1min 行情、[btcusdt, ethbtc] tick 行情
    :param exchange_name: 选择交易所接口 ExchangeName
    :param run_mode: 运行模式 RunMode.Realtime  或 RunMode.Backtest
    :param trade_agent_params: 运行参数，回测模式下：运行起止时间，实时行情下：加载定时器等设置
    :param strategy_handler_param: strategy_handler 运行参数
    :return: 策略执行对象实力
    """
    # 为 md_agent_param 补充参数
    for md_agent_param in md_agent_params_list:
        md_agent_param['exchange_name'] = exchange_name

    # 为 trade_agent_params 补充参数
    trade_agent_params['exchange_name'] = exchange_name
    trade_agent_params['is_default'] = True
    trade_agent_params_list = [trade_agent_params]
    stg_handler = strategy_handler_factory_multi_exchange(
        stg_class, strategy_params, md_agent_params_list, run_mode, trade_agent_params_list, strategy_handler_param)
    return stg_handler


def strategy_handler_factory_multi_exchange(
        stg_class: type(StgBase), strategy_params, md_agent_params_list, run_mode: RunMode,
        trade_agent_params_list: list, strategy_handler_param: dict, add_2_db=True, stg_run_id=None,
        is_4_shown=False) -> StgHandlerBase:
    """
    多交易所策略处理具备
    建立策略对象
    建立数据库相应记录信息
    根据运行模式（实时、回测）：选择相应的md_agent以及trade_agent
    :param stg_class: 策略类型 StgBase 的子类
    :param strategy_params: 策略参数
    :param md_agent_params_list: 行情代理（md_agent）参数，支持同时订阅多周期、多品种，
    例如：同时订阅 [ethusdt, eosusdt] 1min 行情、[btcusdt, ethbtc] tick 行情
    :param run_mode: 运行模式 RunMode.Realtime  或 RunMode.Backtest
    :param trade_agent_params_list: 运行参数，回测模式下：运行起止时间，实时行情下：加载定时器等设置
    :param strategy_handler_param: strategy_handler 运行参数
    :param add_2_db: 默认情况下，自动将 stg_run_info 添加到数据库
    :param stg_run_id: 仅在 add_2_db == False 时需要添加 stg_run_id
    :param is_4_shown: 仅供回溯展示需要，不加载 strategy_class 实例及其相关方法
    :return: 策略执行对象实力
    """
    stg_module = get_module_path(stg_class)
    stg_run_info = StgRunInfo(stg_name=stg_class.__name__,  # 例如："MACrossStg"
                              stg_module=stg_module,  # 例如："ibats_common.example.ma_cross_stg"
                              dt_from=datetime.now(),
                              # dt_to=None,
                              stg_params=json.dumps(strategy_params),
                              md_agent_params_list=json.dumps(md_agent_params_list),
                              run_mode=int(run_mode),
                              trade_agent_params_list=json.dumps(trade_agent_params_list),
                              strategy_handler_param=json.dumps(strategy_handler_param),
                              )
    if add_2_db:
        with with_db_session(engine_ibats) as session:
            session.add(stg_run_info)
            session.commit()
            stg_run_id = stg_run_info.stg_run_id

    # 初始化策略实体，传入参数
    if is_4_shown:
        stg_base = StgBase(**strategy_params)
    else:
        stg_base = stg_class(**strategy_params)
    # 设置相关属性
    stg_base.stg_run_id = stg_run_id
    logger.debug('strategy_params: %s', strategy_params)
    # 设置策略交易接口 trade_agent，这里不适用参数传递的方式而使用属性赋值，
    # 因为stg_base子类被继承后，参数主要用于设置策略所需各种参数使用
    for num, params in enumerate(trade_agent_params_list, start=1):
        logger.debug('%d) run_mode=%s, stg_run_id=%d, trade_agent_params: %s', num, run_mode.name, stg_run_id, params)
        if run_mode == RunMode.Backtest_FixPercent and params['calc_mode'] == CalcMode.Margin:
            raise ValueError(f'{num}) {RunMode.Backtest_FixPercent.name} 模式下，trade_agent 不支持 CalcMode.Margin 模式，'
                             f'请切换到 CalcMode.Normal 模式')
        # 默认使用交易所名称，若同一交易所，多个账户交易，则可以单独指定名称
        agent_name = params.setdefault('agent_name', params['exchange_name'])
        trade_agent = trader_agent_factory(run_mode, stg_run_id, **params)
        stg_base.trade_agent_dic[agent_name] = trade_agent
        if 'is_default' in params and params['is_default']:
            stg_base.trade_agent = trade_agent
            stg_base.trade_agent_dic[ExchangeName.Default] = trade_agent

    # 对不同周期设置相应的md_agent
    # 初始化各个周期的 md_agent
    md_key_period_agent_dic = defaultdict(dict)
    for num, params in enumerate(md_agent_params_list, start=1):
        period = params['md_period']
        agent_name = params.setdefault('agent_name', f"{params['exchange_name']}.{period}")
        logger.debug('%d) run_mode=%s, md_agent_params: %s', num, run_mode.name, params)
        md_agent = md_agent_factory(run_mode=run_mode, **params)
        md_key_period_agent_dic[agent_name][period] = md_agent
        # 对各个周期分别加载历史数据，设置对应 handler
        # 通过 md_agent 加载各个周期的历史数据,
        # 这里加载历史数据为初始化数据：
        # 主要对应于 md_agent_params_list 参数中 init_md_date_from init_md_date_to 等参数设置
        # 与下方的另一次加载历史数据不同，下面的加载历史数据位回测过程中对回测数据的加载，两者不可合并
        his_df_dic = md_agent.load_history()
        if his_df_dic is None:
            logger.warning('加载 %s 历史数据为 None', period)
            continue
        if isinstance(his_df_dic, dict):
            md_df = his_df_dic['md_df']
        else:
            md_df = his_df_dic
            warnings.warn('load_history 返回 df 数据格式即将废弃，请更新成 dict', DeprecationWarning)

        context = {
            ContextKey.instrument_id_list: list(md_agent.instrument_id_list),
            ContextKey.md_agent_key: agent_name,
            ContextKey.period: period,
        }
        stg_base.load_md_period_df(period, md_df, context)
        logger.debug('加载 %s 历史数据 %s 条', period, 'None' if md_df is None else str(md_df.shape[0]))

    # 设置 md_td_agent_key_list_map 用于标识 md_agent_key 与 trade_agent_key 之间的对应关系
    # md_agent_key 与 trade_agent_key 是 1对多 关系
    if 'md_td_agent_key_list_map' in strategy_handler_param:
        md_td_agent_key_list_map = strategy_handler_param['md_td_agent_key_list_map']
    else:
        md_td_agent_key_list_map = defaultdict(list)
        # 遍历所有 md_agent, 查找 exchange_name 相同的进行 mapping
        for md_agent_key, period_agent_dic in md_key_period_agent_dic.items():
            for period, md_agent in period_agent_dic.items():
                for trade_agent_key, trade_agent in stg_base.trade_agent_dic.items():
                    if md_agent.exchange_name == trade_agent.exchange_name:
                        md_td_agent_key_list_map[md_agent_key].append(trade_agent_key)

                break
        # 整理 mapping 去除重复 key，整理过程中 td_agent_key_list 顺序可能被打乱
        for md_agent_key in list(md_td_agent_key_list_map.keys()):
            td_agent_key_list = list(set(md_td_agent_key_list_map[md_agent_key]))
            md_td_agent_key_list_map[md_agent_key] = td_agent_key_list
        strategy_handler_param['md_td_agent_key_list_map'] = md_td_agent_key_list_map
    # TODO: 可以增加一个合法性检查，检查是否 md_agent_key 与 trade_agent_key 是 1对多 关系
    logger.debug('md_td_agent_key_list_map: %s', md_td_agent_key_list_map)

    # 为 stg_base 设置 md_td_agent_key_list_map
    stg_base.set_md_td_agent_key_list_map(md_td_agent_key_list_map)

    # 初始化 StgHandlerBase 实例
    logger.debug("stg_run_id=%d,\n\tstrategy_handler_param: %s", stg_run_id, strategy_handler_param)
    if run_mode == RunMode.Realtime:
        stg_handler = StgHandlerRealtime(
            stg_run_id=stg_run_id, stg_base=stg_base, md_key_period_agent_dic=md_key_period_agent_dic,
            **strategy_handler_param)
    elif run_mode in (RunMode.Backtest, RunMode.Backtest_FixPercent):
        stg_handler = StgHandlerBacktest(
            stg_run_id=stg_run_id, stg_base=stg_base, run_mode=run_mode,
            md_key_period_agent_dic=md_key_period_agent_dic,
            **strategy_handler_param)
    else:
        raise ValueError('run_mode %d error' % run_mode)

    logger.debug('初始化 %r 完成', stg_handler)
    return stg_handler


@lru_cache()
def strategy_handler_loader(stg_run_id, is_4_shown=True) -> StgHandlerBase:
    """
    根据 stg_run_id 从数据库中加载相应参数，动态加载策略类，实例化 stg_handler 类
    :param stg_run_id:
    :param is_4_shown: 仅供回溯展示需要，不加载 strategy_class 实例及其相关方法
    :return:
    """
    with with_db_session(engine_ibats) as session:
        stg_run_info = session.query(StgRunInfo).filter(StgRunInfo.stg_run_id == stg_run_id).first()
    if stg_run_info is None:
        raise KeyError(f'stg_run_id={stg_run_id} 无效，数据库中没有相关数据')
    # 2019-05-14 废弃参数 module_name_replacement_if_main
    # 因版本升级该问题以及被解决，因此该参数已经无用
    # if module_name_replacement_if_main is not None and stg_run_info.stg_module == '__main__':
    #     # stg_run_info.stg_module = "ibats_common.example.ma_cross_stg"
    #     stg_run_info.stg_module = module_name_replacement_if_main

    stg_class = load_class(stg_run_info.stg_module, stg_run_info.stg_name)

    strategy_params = json.loads(stg_run_info.stg_params)
    md_agent_params_list = json.loads(stg_run_info.md_agent_params_list)
    run_mode = RunMode(stg_run_info.run_mode)
    trade_agent_params_list = json.loads(stg_run_info.trade_agent_params_list)
    strategy_handler_param = json.loads(stg_run_info.strategy_handler_param)
    stg_handler = strategy_handler_factory_multi_exchange(
        stg_class=stg_class,
        strategy_params=strategy_params,
        md_agent_params_list=md_agent_params_list,
        run_mode=run_mode,
        trade_agent_params_list=trade_agent_params_list,
        strategy_handler_param=strategy_handler_param,
        add_2_db=False,
        stg_run_id=stg_run_info.stg_run_id,
        is_4_shown=is_4_shown,
    )
    return stg_handler



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 19-5-9 下午4:34
@File    : plot_db.py
@contact : mmmaaaggg@163.com
@desc    : 需要依赖于数据库进行输出展示的函数
"""
import logging
import os
from collections import defaultdict

import matplotlib.pyplot as plt
import pandas as pd
from ibats_utils.db import with_db_session
from ibats_utils.mess import date_2_str
from sqlalchemy.sql import func

from ibats_common.analysis import get_cache_folder_path
from ibats_common.analysis.plot import get_file_name
from ibats_common.backend import engines
from ibats_common.backend.mess import get_stg_run_id_latest
from ibats_common.backend.orm import StgRunStatusDetail, OrderDetail, TradeDetail, StgRunInfo
from ibats_common.common import Action, Direction, RunMode
from ibats_common.strategy_handler import strategy_handler_loader

logger = logging.getLogger(__name__)


def show_cash_and_margin(stg_run_id, enable_show_plot=True, enable_save_plot=False, run_mode=RunMode.Backtest, **kwargs):
    """
    plot cash_and_margin
    :param stg_run_id:
    :param enable_show_plot:
    :param enable_save_plot:
    :param run_mode:
    :param kwargs:
    :return:
    """
    # stg_run_id=154
    engine_ibats = engines.engine_ibats
    with with_db_session(engine_ibats) as session:
        if stg_run_id is None:
            stg_run_id = session.query(func.max(StgRunInfo.stg_run_id)).scalar()
            logger.warning('没有设置 stg_run_id 参数，将输出最新的 stg_run_id=%d 对应记录', stg_run_id)

        sql_str = str(
            session.query(
                StgRunStatusDetail.trade_dt.label('trade_dt'),
                StgRunStatusDetail.cash_available.label('cash'),
                (StgRunStatusDetail.cash_available.label('cash') + StgRunStatusDetail.commission_tot
                 ).label('cash + commission'),
                StgRunStatusDetail.curr_margin.label('margin'),
                StgRunStatusDetail.cash_and_margin.label('cash_and_margin'),
                (StgRunStatusDetail.cash_and_margin.label('cash_and_margin') + StgRunStatusDetail.commission_tot
                 ).label('no commission'),
            ).filter(
                StgRunStatusDetail.stg_run_id == stg_run_id
            )
        )

    df = pd.read_sql(sql_str, engine_ibats, params=[stg_run_id], index_col=['trade_dt'])

    def plot_func():
        ax = df[['cash', 'margin']].plot.area()
        if run_mode != RunMode.Backtest_FixPercent:
            df[['cash_and_margin', 'no commission', 'cash + commission']].plot(ax=ax)

        ax.set_title(
            f"Cash + Margin [{stg_run_id}] "
            f"{date_2_str(min(df.index))} - {date_2_str(max(df.index))} ({df.shape[0]} days)")

    if enable_show_plot:
        plot_func()
        plt.show()

    if enable_save_plot:
        plot_func()
        file_name = get_file_name(f'cash_and_margin', name=stg_run_id)
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    return df, file_path


def get_md(stg_run_id, return_td_md_agent_key_map=False):
    # 获取行情数据
    stg_handler = strategy_handler_loader(stg_run_id, is_4_shown=True)

    td_md_agent_key_map = stg_handler.stg_base._td_md_agent_key_map

    # 获取历史行情数据
    md_agent_key_cor_func_dic = stg_handler.get_periods_history_iterator()
    # 根据 md_agent 对每一组行情 以及 对应的 order_detail_list 进行 plot
    # fig = plt.figure(1, figsize=(20, 4.8 * agent_count))
    symbol_rr_dic = {}
    sum_rr_df, sum_col_name_list = None, []
    for num, ((md_agent_key, period), (cor_func, meta_dic)) in enumerate(md_agent_key_cor_func_dic.items(), start=1):
        df = pd.DataFrame([md_s for num, datetime_tag, md_s in cor_func])
        if df.shape[0] == 0:
            continue
        # ax = fig.add_subplot(num, 1, 1)
        # 行情
        symbol_key = meta_dic['symbol_key']
        close_key = meta_dic['close_key']
        timestamp_key = meta_dic['timestamp_key']
        df.set_index(timestamp_key, inplace=True)
        for symbol, df_by_symbol in df.groupby(symbol_key):
            symbol_rr_dic[(md_agent_key, period, symbol)] = (df_by_symbol, close_key)

            # 汇总输出
            col_name = f"{symbol}_rr"
            sum_col_name_list.append(col_name)
            md_df = df_by_symbol[[close_key]].copy()
            md_df['md_rr'] = md_df[close_key] / md_df[close_key].iloc[0]
            if sum_rr_df is None:
                sum_rr_df = md_df.rename(columns={'md_rr': col_name})[sum_col_name_list]
            else:
                sum_rr_df = sum_rr_df.join(md_df.rename(columns={'md_rr': col_name}))[sum_col_name_list]

    if return_td_md_agent_key_map:
        return sum_rr_df, symbol_rr_dic, td_md_agent_key_map
    else:
        return sum_rr_df, symbol_rr_dic


def get_rr_with_md(stg_run_id, compound_rr=True):
    """
    获取策略收益率数据
    :param stg_run_id:
    :param compound_rr:复合收益率
    :return:
    """
    engine_ibats = engines.engine_ibats
    # 获取 收益曲线
    with with_db_session(engine_ibats) as session:
        if compound_rr:
            sql_str = str(
                session.query(
                    StgRunStatusDetail.trade_dt.label('trade_dt'),
                    StgRunStatusDetail.cash_and_margin.label('cash and margin'),
                    (StgRunStatusDetail.cash_and_margin.label('cash_and_margin') +
                     StgRunStatusDetail.commission_tot.label('commission_tot')).label('no_commission'),
                    StgRunStatusDetail.rr_compound.label('rr'),
                    StgRunStatusDetail.rr_compound_nc.label('rr no commission'),
                ).filter(
                    StgRunStatusDetail.stg_run_id == stg_run_id
                )
            )
        else:
            sql_str = str(
                session.query(
                    StgRunStatusDetail.trade_dt.label('trade_dt'),
                    StgRunStatusDetail.cash_and_margin.label('cash and margin'),
                    (StgRunStatusDetail.cash_and_margin.label('cash_and_margin') +
                     StgRunStatusDetail.commission_tot.label('commission_tot')).label('no_commission'),
                    StgRunStatusDetail.rr.label('rr'),
                    StgRunStatusDetail.rr_nc.label('rr no commission'),
                ).filter(
                    StgRunStatusDetail.stg_run_id == stg_run_id
                )
            )

    rr_df = pd.read_sql(sql_str, engine_ibats, params=[stg_run_id]).set_index('trade_dt')
    if rr_df is None or rr_df.shape[0] == 0:
        return None, None

    # rr_df['rr'] = rr_df['cash and margin'] / rr_df['cash and margin'].iloc[0]
    # rr_df['rr without commission'] = rr_df['without commission'] / rr_df['without commission'].iloc[0]
    col_list_rr = ['rr', 'rr no commission']
    rr_df[col_list_rr] += 1

    # 获取行情数据
    sum_df, symbol_rr_dic = get_md(stg_run_id)
    sum_df = sum_df.join(rr_df[col_list_rr])

    col_list = ['md_rr']
    col_list.extend(col_list)
    for num, (key, (df, close_key)) in enumerate(symbol_rr_dic.items()):
        md_df = df[[close_key]].copy()
        md_df['md_rr'] = md_df[close_key] / md_df[close_key].iloc[0]
        md_df = md_df.join(rr_df)[col_list]
        symbol_rr_dic[key] = (md_df, close_key)

    return sum_df, symbol_rr_dic


def _test_show_rr_with_md():
    stg_run_id = None
    show_rr_with_md(stg_run_id)


def show_rr_with_md(stg_run_id, show_sum_plot=True, show_each_md_plot=False, enable_save_plot=False):
    if stg_run_id is None:
        stg_run_id = get_stg_run_id_latest()

    sum_df, symbol_rr_dic = get_rr_with_md(stg_run_id)
    save_file_path_dic = {}

    if show_each_md_plot:
        for num, (symbol, (df, close_key)) in enumerate(symbol_rr_dic.items(), start=1):
            ax = df.plot()
            ax.set_title(
                f"Return Rate {symbol} [{stg_run_id}] "
                f"{date_2_str(min(df.index))} - {date_2_str(max(df.index))} ({df.shape[0]} days)")
            plt.show()

    if show_sum_plot:
        ax = sum_df.plot()
        ax.set_title(
            f"Return Rate [{stg_run_id}] "
            f"{date_2_str(min(sum_df.index))} - {date_2_str(max(sum_df.index))} ({sum_df.shape[0]} days)")
        plt.show()

    if enable_save_plot:
        ax = sum_df.plot()
        ax.set_title(
            f"Return Rate [{stg_run_id}] "
            f"{date_2_str(min(sum_df.index))} - {date_2_str(max(sum_df.index))} ({sum_df.shape[0]} days)")

        rr_plot_file_path = os.path.join(get_cache_folder_path(), f'rr_plot {stg_run_id}.png')
        plt.savefig(rr_plot_file_path, dpi=75)
        save_file_path_dic['rr'] = rr_plot_file_path

    return sum_df, symbol_rr_dic, save_file_path_dic


def show_order(stg_run_id, **kwargs) -> (defaultdict(lambda: defaultdict(list)), str):
    """
    plot candle and buy and sell point
    :param stg_run_id:
    :param kwargs:
    :return:
    """
    # 加载数据库 engine
    engine_ibats = engines.engine_ibats
    # stg_run_id = 1
    if stg_run_id is None:
        stg_run_id = get_stg_run_id_latest()

    # 获取行情数据
    sum_rr_df, symbol_rr_dic, td_md_agent_key_map = get_md(stg_run_id, return_td_md_agent_key_map=True)

    # 获取全部订单
    # session = get_db_session(engine_ibats)
    with with_db_session(engine_ibats) as session:
        order_detail_list_tot = session.query(
            OrderDetail
            # OrderDetail.order_dt.label('order_dt'),
            # OrderDetail.order_price.label('order_price'),
            # OrderDetail.action.label('action'),
            # OrderDetail.direction.label('direction'),
            # OrderDetail.order_vol.label('order_vol'),
            # OrderDetail.symbol.label('symbol'),
            # OrderDetail.trade_agent_key.label('trade_agent_key'),
        ).filter(
            OrderDetail.stg_run_id == stg_run_id
        ).all()

    # 根据 md_agent 进行分组
    md_agent_key_order_detail_list_dic = defaultdict(list)
    for num, detail in enumerate(order_detail_list_tot):
        md_agent_key = td_md_agent_key_map[detail.trade_agent_key]
        md_agent_key_order_detail_list_dic[md_agent_key].append(detail)

    # 获取历史行情数据
    data_dict = defaultdict(lambda: defaultdict(list))
    for num, ((md_agent_key, period, symbol), (df_by_symbol, close_key)) in enumerate(symbol_rr_dic.items(), start=1):
        detail_list = md_agent_key_order_detail_list_dic[md_agent_key]
        data_dict[(md_agent_key, period, symbol)]['md'].append(df_by_symbol[close_key])
        # 开仓
        order_detail_list_sub = [_ for _ in detail_list
                                 if _.symbol == symbol
                                 and ((_.direction == Direction.Long.value and _.action == Action.Open.value)
                                      or (_.direction == Direction.Short.value and _.action != Action.Open.value)
                                      )
                                 ]
        trade_date_list = [_.order_dt for _ in order_detail_list_sub]
        price = [_.order_price for _ in order_detail_list_sub]
        # ax.scatter(trade_date_list, price, c='r', marker='^')
        data_dict[(md_agent_key, period, symbol)]['long_open_or_short_close'].append((trade_date_list, price))
        # 关仓
        order_detail_list_sub = [_ for _ in detail_list
                                 if (_.direction == Direction.Long.value and _.action != Action.Open.value)
                                 or (_.direction == Direction.Short.value and _.action == Action.Open.value)]
        trade_date_list = [_.order_dt for _ in order_detail_list_sub]
        price = [_.order_price for _ in order_detail_list_sub]
        # ax.scatter(trade_date_list, price, c='g', marker='v')
        data_dict[(md_agent_key, period, symbol)]['short_open_or_long_close'].append((trade_date_list, price))
        # 建立连线
        order_detail_list_symbol = [_ for _ in detail_list if _.symbol == symbol]
        for point1, point2 in zip(order_detail_list_symbol[:-1], order_detail_list_symbol[1:]):
            if point1.order_dt == point2.order_dt:
                # logger.debug("%s %f %s ignore", point2.order_dt, point2.order_price, point2.action)
                continue
            # logger.debug("%s %f -> %s %f %d",
            #              point1.order_dt, point1.order_price, point2.order_dt, point2.order_price, point2.action)
            # ax.plot([point1.order_dt, point2.order_dt], [point1.order_price, point2.order_price],
            #         c='r' if point2.direction != Direction.Long.value else 'g')
            if point2.direction != Direction.Long:
                data_dict[(md_agent_key, period, symbol)]['buy_sell_point_pair'].append(
                    ([point1.order_dt, point2.order_dt], [point1.order_price, point2.order_price])
                )
            else:
                data_dict[(md_agent_key, period, symbol)]['sell_buy_point_pair'].append(
                    ([point1.order_dt, point2.order_dt], [point1.order_price, point2.order_price])
                )

    # show
    file_path = show_plot_data_dic(data_dict, title=f"MD and Order figure [{stg_run_id}]", **kwargs)
    return data_dict, file_path


def show_trade(stg_run_id, **kwargs) -> (defaultdict(lambda: defaultdict(list)), str):
    """
    plot candle and buy and sell point
    :param stg_run_id:
    :param kwargs:
    :return:
    """
    # stg_run_id = 1
    if stg_run_id is None:
        stg_run_id = get_stg_run_id_latest()

    # 获取行情数据
    sum_rr_df, symbol_rr_dic, td_md_agent_key_map = get_md(stg_run_id, return_td_md_agent_key_map=True)

    # 加载数据库 engine
    engine_ibats = engines.engine_ibats
    # 获取全部订单
    # session = get_db_session(engine_ibats)
    with with_db_session(engine_ibats) as session:
        data_list_tot = session.query(
            TradeDetail
        ).filter(
            TradeDetail.stg_run_id == stg_run_id
        ).all()

    # 根据 md_agent 进行分组
    md_agent_key_order_detail_list_dic = defaultdict(list)
    for num, detail in enumerate(data_list_tot):
        md_agent_key = td_md_agent_key_map[detail.trade_agent_key]
        md_agent_key_order_detail_list_dic[md_agent_key].append(detail)

    # 获取历史行情数据
    data_dict = defaultdict(lambda: defaultdict(list))
    for num, ((md_agent_key, period, symbol), (df_by_symbol, close_key)) in enumerate(symbol_rr_dic.items(), start=1):
        detail_list = md_agent_key_order_detail_list_dic[md_agent_key]
        data_dict[(md_agent_key, period, symbol)]['md'].append(df_by_symbol[close_key])
        # 开仓
        detail_list_sub = [_ for _ in detail_list
                           if _.symbol == symbol
                           and ((_.direction == Direction.Long.value and _.action == Action.Open.value)
                                or (_.direction == Direction.Short.value and _.action != Action.Open.value)
                                )
                           ]
        trade_date_list = [_.trade_dt for _ in detail_list_sub]
        price = [_.trade_price for _ in detail_list_sub]
        # ax.scatter(trade_date_list, price, c='r', marker='^')
        data_dict[(md_agent_key, period, symbol)]['long_open_or_short_close'].append((trade_date_list, price))
        # 关仓
        detail_list_sub = [_ for _ in detail_list
                           if (_.direction == Direction.Long.value and _.action != Action.Open.value)
                           or (_.direction == Direction.Short.value and _.action == Action.Open.value)]
        trade_date_list = [_.trade_dt for _ in detail_list_sub]
        price = [_.trade_price for _ in detail_list_sub]
        # ax.scatter(trade_date_list, price, c='g', marker='v')
        data_dict[(md_agent_key, period, symbol)]['short_open_or_long_close'].append((trade_date_list, price))
        # 建立连线
        detail_list_symbol = [_ for _ in detail_list if _.symbol == symbol]
        for point1, point2 in zip(detail_list_symbol[:-1], detail_list_symbol[1:]):
            if point1.trade_dt == point2.trade_dt:
                # logger.debug("%s %f %s ignore", point2.order_dt, point2.order_price, point2.action)
                continue
            # logger.debug("%s %f -> %s %f %d",
            #              point1.order_dt, point1.order_price, point2.order_dt, point2.order_price, point2.action)
            # ax.plot([point1.order_dt, point2.order_dt], [point1.order_price, point2.order_price],
            #         c='r' if point2.direction != Direction.Long.value else 'g')
            if point2.direction != Direction.Long:
                data_dict[(md_agent_key, period, symbol)]['buy_sell_point_pair'].append(
                    ([point1.trade_dt, point2.trade_dt], [point1.trade_price, point2.trade_price])
                )
            else:
                data_dict[(md_agent_key, period, symbol)]['sell_buy_point_pair'].append(
                    ([point1.trade_dt, point2.trade_dt], [point1.trade_price, point2.trade_price])
                )

    # show
    file_path = show_plot_data_dic(data_dict, title=f"MD and Trade figure [{stg_run_id}]", **kwargs)
    return data_dict, file_path


def show_plot_data_dic(data_dict: dict, title=None, enable_show_plot=True, enable_save_plot=False, **kwargs):
    """
    将数据plot展示出来
    :param data_dict:
    :param title:
    :param enable_show_plot:
    :param enable_save_plot:
    :param kwargs:
    :return:
    """
    data_len = len(data_dict)

    def func():
        fig, axs = plt.subplots(
            data_len, 1,
            # constrained_layout=True,
            figsize=(20, 4.8 * data_len))
        name = '' if title is None else title

        # 容易与 ax title 重叠
        # if title is not None:
        #     fig.suptitle(title, fontsize=16)
        for num, ((md_agent_key, period, symbol), plot_data_dic) in enumerate(data_dict.items()):
            # ax = fig.add_subplot(num, 1, 1)
            ax = axs[num] if data_len > 1 else axs
            ax.set_title(f"{name} - md_agent_key={md_agent_key} - period={period} - symbol={symbol}")
            for md in plot_data_dic['md']:
                md.plot(ax=ax, colormap='jet')
            for x, y in plot_data_dic['long_open_or_short_close']:
                ax.scatter(x, y, c='r', marker='^')
            for x, y in plot_data_dic['short_open_or_long_close']:
                ax.scatter(x, y, c='g', marker='v')
            for x, y in plot_data_dic['buy_sell_point_pair']:
                ax.plot(x, y, c='r')
            for x, y in plot_data_dic['sell_buy_point_pair']:
                ax.plot(x, y, c='g')

    if enable_show_plot:
        func()
        plt.show()

    if enable_save_plot:
        func()
        file_name = get_file_name(f'title', name=title)
        file_path = os.path.join(get_cache_folder_path(), file_name)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    return file_path


def _test_show_rr_with_md():
    stg_run_id = None
    # data_dict = show_order(
    #     stg_run_id,
    #     # module_name_replacement_if_main='ibats_common.example.tf_stg.ai_stg',
    #     module_name_replacement_if_main='ibats_common.example.ma_cross_stg',
    # )
    # show_cash_and_margin(stg_run_id)
    show_rr_with_md(stg_run_id, show_each_md_plot=True, show_sum_plot=True, enable_save_plot=True)


def _test_show_cash_and_margin():
    stg_run_id = None
    show_cash_and_margin(stg_run_id, enable_show_plot=True, enable_save_plot=True)


if __name__ == '__main__':
    # _test_show_rr_with_md()
    # _test_show_rr_with_md()
    _test_show_cash_and_margin()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/26 10:27
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from ibats_utils.mess import get_folder_path
import os
import re


module_root_path = get_folder_path(re.compile(r'^ibats[\w]+'), create_if_not_found=False)  # 'ibats_common'
root_parent_path = os.path.abspath(
    os.path.join(module_root_path, os.path.pardir))
